# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functions for building code during presubmit checks."""

import base64
import contextlib
from dataclasses import dataclass
import itertools
import json
import logging
import os
import posixpath
from pathlib import Path
import re
import subprocess
from shutil import which
import sys
import tarfile
from typing import (
    Any,
    Callable,
    Collection,
    Container,
    ContextManager,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
    Set,
    TextIO,
)

import pw_cli.color
from pw_cli.plural import plural
from pw_cli.file_filter import FileFilter
from pw_presubmit.presubmit import (
    call,
    Check,
    filter_paths,
    install_package,
    PresubmitResult,
    SubStep,
)
from pw_presubmit.presubmit_context import (
    PresubmitContext,
    PresubmitFailure,
)
from pw_presubmit import (
    bazel_parser,
    format_code,
    ninja_parser,
)
from pw_presubmit.tools import (
    log_run,
    format_command,
)

_LOG = logging.getLogger(__name__)


BAZEL_EXECUTABLE = 'bazel'


def bazel(
    ctx: PresubmitContext,
    cmd: str,
    *args: str,
    strict_module_lockfile: bool = False,
    use_remote_cache: bool = False,
    stdout: TextIO | None = None,
    **kwargs,
) -> None:
    """Invokes Bazel with some common flags set.

    Intended for use with bazel build and test. May not work with others.
    """

    num_jobs: list[str] = []
    if ctx.num_jobs is not None:
        num_jobs.extend(('--jobs', str(ctx.num_jobs)))

    keep_going: list[str] = []
    if ctx.continue_after_build_error:
        keep_going.append('--keep_going')

    strict_lockfile: list[str] = []
    if strict_module_lockfile:
        strict_lockfile.append('--lockfile_mode=error')

    remote_cache: list[str] = []
    if use_remote_cache and ctx.luci:
        remote_cache.append('--config=remote_cache')
        if ctx.luci.is_ci:
            # Only CI builders should attempt to write to the cache. Try
            # builders will be denied permission if they do so.
            remote_cache.append('--remote_upload_local_results=true')

    ctx.output_dir.mkdir(exist_ok=True, parents=True)
    try:
        with contextlib.ExitStack() as stack:
            if not stdout:
                stdout = stack.enter_context(
                    (ctx.output_dir / f'bazel.{cmd}.stdout').open('w')
                )

            with (ctx.output_dir / 'bazel.output.base').open('w') as outs, (
                ctx.output_dir / 'bazel.output.base.err'
            ).open('w') as errs:
                call(
                    BAZEL_EXECUTABLE,
                    'info',
                    'output_base',
                    tee=outs,
                    stderr=errs,
                )

            call(
                BAZEL_EXECUTABLE,
                cmd,
                '--verbose_failures',
                '--worker_verbose',
                f'--symlink_prefix={ctx.output_dir / "bazel-"}',
                *num_jobs,
                *keep_going,
                *strict_lockfile,
                *remote_cache,
                *args,
                cwd=ctx.root,
                tee=stdout,
                call_annotation={'build_system': 'bazel'},
                **kwargs,
            )

    except PresubmitFailure as exc:
        if stdout:
            failure = bazel_parser.parse_bazel_stdout(Path(stdout.name))
            if failure:
                with ctx.failure_summary_log.open('w') as outs:
                    outs.write(failure)

        raise exc


def _gn_value(value) -> str:
    if isinstance(value, bool):
        return str(value).lower()

    if (
        isinstance(value, str)
        and '"' not in value
        and not value.startswith("{")
        and not value.startswith("[")
    ):
        return f'"{value}"'

    if isinstance(value, (list, tuple)):
        return f'[{", ".join(_gn_value(a) for a in value)}]'

    # Fall-back case handles integers as well as strings that already
    # contain double quotation marks, or look like scopes or lists.
    return str(value)


def gn_args_list(**kwargs) -> list[str]:
    """Return a list of formatted strings to use as gn args.

    Currently supports bool, int, and str values. In the case of str values,
    quotation marks will be added automatically, unless the string already
    contains one or more double quotation marks, or starts with a { or [
    character, in which case it will be passed through as-is.
    """
    transformed_args = []
    for arg, val in kwargs.items():
        transformed_args.append(f'{arg}={_gn_value(val)}')

    # Use ccache if available for faster repeat presubmit runs.
    if which('ccache') and 'pw_command_launcher' not in kwargs:
        transformed_args.append('pw_command_launcher="ccache"')

    return transformed_args


def gn_args(**kwargs) -> str:
    """Builds a string to use for the --args argument to gn gen.

    Currently supports bool, int, and str values. In the case of str values,
    quotation marks will be added automatically, unless the string already
    contains one or more double quotation marks, or starts with a { or [
    character, in which case it will be passed through as-is.
    """
    return '--args=' + ' '.join(gn_args_list(**kwargs))


def write_gn_args_file(destination_file: Path, **kwargs) -> str:
    """Write gn args to a file.

    Currently supports bool, int, and str values. In the case of str values,
    quotation marks will be added automatically, unless the string already
    contains one or more double quotation marks, or starts with a { or [
    character, in which case it will be passed through as-is.

    Returns:
      The contents of the written file.
    """
    contents = '\n'.join(gn_args_list(**kwargs))
    # Add a trailing linebreak
    contents += '\n'
    destination_file.parent.mkdir(exist_ok=True, parents=True)

    if (
        destination_file.is_file()
        and destination_file.read_text(encoding='utf-8') == contents
    ):
        # File is identical, don't re-write.
        return contents

    destination_file.write_text(contents, encoding='utf-8')
    return contents


def gn_gen(
    ctx: PresubmitContext,
    *args: str,
    gn_check: bool = True,  # pylint: disable=redefined-outer-name
    gn_fail_on_unused: bool = True,
    export_compile_commands: bool | str = True,
    preserve_args_gn: bool = False,
    **gn_arguments,
) -> None:
    """Runs gn gen in the specified directory with optional GN args.

    Runs with --check=system if gn_check=True. Note that this does not cover
    generated files. Run gn_check() after building to check generated files.
    """
    all_gn_args = {'pw_build_COLORIZE_OUTPUT': pw_cli.color.is_enabled()}
    all_gn_args.update(gn_arguments)
    all_gn_args.update(ctx.override_gn_args)
    _LOG.debug('%r', all_gn_args)
    args_option = gn_args(**all_gn_args)

    if not ctx.dry_run and not preserve_args_gn:
        # Delete args.gn to ensure this is a clean build.
        args_gn = ctx.output_dir / 'args.gn'
        if args_gn.is_file():
            args_gn.unlink()

    export_commands_arg = ''
    if export_compile_commands:
        export_commands_arg = '--export-compile-commands'
        if isinstance(export_compile_commands, str):
            export_commands_arg += f'={export_compile_commands}'

    call(
        'gn',
        '--color' if pw_cli.color.is_enabled() else '--nocolor',
        'gen',
        ctx.output_dir,
        *(['--check=system'] if gn_check else []),
        *(['--fail-on-unused-args'] if gn_fail_on_unused else []),
        *([export_commands_arg] if export_commands_arg else []),
        *args,
        *([args_option] if all_gn_args else []),
        cwd=ctx.root,
        call_annotation={
            'gn_gen_args': all_gn_args,
            'gn_gen_args_option': args_option,
        },
    )


def gn_check(ctx: PresubmitContext) -> PresubmitResult:
    """Runs gn check, including on generated and system files."""
    call(
        'gn',
        'check',
        ctx.output_dir,
        '--check-generated',
        '--check-system',
        cwd=ctx.root,
    )
    return PresubmitResult.PASS


def ninja(
    ctx: PresubmitContext,
    *args,
    save_compdb: bool = True,
    save_graph: bool = True,
    **kwargs,
) -> None:
    """Runs ninja in the specified directory."""

    num_jobs: list[str] = []
    if ctx.num_jobs is not None:
        num_jobs.extend(('-j', str(ctx.num_jobs)))

    keep_going: list[str] = []
    if ctx.continue_after_build_error:
        keep_going.extend(('-k', '0'))

    if save_compdb:
        proc = log_run(
            ['ninja', '-C', ctx.output_dir, '-t', 'compdb', *args],
            capture_output=True,
            **kwargs,
        )
        if not ctx.dry_run:
            (ctx.output_dir / 'ninja.compdb').write_bytes(proc.stdout)

    if save_graph:
        proc = log_run(
            ['ninja', '-C', ctx.output_dir, '-t', 'graph', *args],
            capture_output=True,
            **kwargs,
        )
        if not ctx.dry_run:
            (ctx.output_dir / 'ninja.graph').write_bytes(proc.stdout)

    ninja_stdout = ctx.output_dir / 'ninja.stdout'
    ctx.output_dir.mkdir(exist_ok=True, parents=True)
    try:
        with ninja_stdout.open('w') as outs:
            if sys.platform == 'win32':
                # Windows doesn't support pw-wrap-ninja.
                ninja_command = ['ninja']
            else:
                ninja_command = ['pw-wrap-ninja', '--log-actions']

            call(
                *ninja_command,
                '-C',
                ctx.output_dir,
                *num_jobs,
                *keep_going,
                *args,
                tee=outs,
                propagate_sigterm=True,
                call_annotation={'build_system': 'ninja'},
                **kwargs,
            )

    except PresubmitFailure as exc:
        failure = ninja_parser.parse_ninja_stdout(ninja_stdout)
        if failure:
            with ctx.failure_summary_log.open('w') as outs:
                outs.write(failure)

        raise exc


def get_gn_args(directory: Path) -> list[dict[str, dict[str, str]]]:
    """Dumps GN variables to JSON."""
    proc = log_run(
        ['gn', 'args', directory, '--list', '--json'], stdout=subprocess.PIPE
    )
    return json.loads(proc.stdout)


def cmake(
    ctx: PresubmitContext,
    *args: str,
    env: Mapping['str', 'str'] | None = None,
) -> None:
    """Runs CMake for Ninja on the given source and output directories."""
    call(
        'cmake',
        '-B',
        ctx.output_dir,
        '-S',
        ctx.root,
        '-G',
        'Ninja',
        *args,
        env=env,
    )


def env_with_clang_vars() -> Mapping[str, str]:
    """Returns the environment variables with CC, CXX, etc. set for clang."""
    env = os.environ.copy()
    env['CC'] = env['LD'] = env['AS'] = 'clang'
    env['CXX'] = 'clang++'
    return env


def _get_paths_from_command(source_dir: Path, *args, **kwargs) -> Set[Path]:
    """Runs a command and reads Bazel or GN //-style paths from it."""
    process = log_run(args, capture_output=True, cwd=source_dir, **kwargs)

    if process.returncode:
        _LOG.error(
            'Build invocation failed with return code %d!', process.returncode
        )
        _LOG.error(
            '[COMMAND] %s\n%s\n%s',
            *format_command(args, kwargs),
            process.stderr.decode(),
        )
        raise PresubmitFailure

    files = set()

    for line in process.stdout.splitlines():
        path = line.strip().lstrip(b'/').replace(b':', b'/').decode()
        path = source_dir.joinpath(path)
        if path.is_file():
            files.add(path)

    return files


# Finds string literals with '.' in them.
_MAYBE_A_PATH = re.compile(
    r'"'  # Starting double quote.
    # Start capture group 1 - the whole filename:
    #   File basename, a single period, file extension.
    r'([^\n" ]+\.[^\n" ]+)'
    # Non-capturing group 2 (optional).
    r'(?: > [^\n"]+)?'  # pw_zip style string "input_file.txt > output_file.txt"
    r'"'  # Ending double quote.
)


def _search_files_for_paths(build_files: Iterable[Path]) -> Iterable[Path]:
    for build_file in build_files:
        directory = build_file.parent

        for string in _MAYBE_A_PATH.finditer(build_file.read_text()):
            path = directory / string.group(1)
            if path.is_file():
                yield path


def _read_compile_commands(compile_commands: Path) -> dict:
    with compile_commands.open('rb') as fd:
        return json.load(fd)


def compiled_files(compile_commands: Path) -> Iterable[Path]:
    for command in _read_compile_commands(compile_commands):
        file = Path(command['file'])
        if file.is_absolute():
            yield file
        else:
            yield file.joinpath(command['directory']).resolve()


def check_compile_commands_for_files(
    compile_commands: Path | Iterable[Path],
    files: Iterable[Path],
    extensions: Collection[str] = format_code.CPP_SOURCE_EXTS,
) -> list[Path]:
    """Checks for paths in one or more compile_commands.json files.

    Only checks C and C++ source files by default.
    """
    if isinstance(compile_commands, Path):
        compile_commands = [compile_commands]

    compiled = frozenset(
        itertools.chain.from_iterable(
            compiled_files(cmds) for cmds in compile_commands
        )
    )
    return [f for f in files if f not in compiled and f.suffix in extensions]


def check_bazel_build_for_files(
    bazel_extensions_to_check: Container[str],
    files: Iterable[Path],
    bazel_dirs: Iterable[Path] = (),
) -> list[Path]:
    """Checks that source files are in the Bazel builds.

    Args:
        bazel_extensions_to_check: which file suffixes to look for in Bazel
        files: the files that should be checked
        bazel_dirs: directories in which to run bazel query

    Returns:
        a list of missing files; will be empty if there were no missing files
    """

    # Collect all paths in the Bazel builds.
    bazel_builds: Set[Path] = set()
    for directory in bazel_dirs:
        bazel_builds.update(
            _get_paths_from_command(
                directory,
                BAZEL_EXECUTABLE,
                'query',
                'kind("source file", //...:*)',
            )
        )

    missing: list[Path] = []

    if bazel_dirs:
        for path in (p for p in files if p.suffix in bazel_extensions_to_check):
            if path not in bazel_builds:
                # TODO: b/234883555 - Replace this workaround for fuzzers.
                if 'fuzz' not in str(path):
                    missing.append(path)

    if missing:
        _LOG.warning(
            '%s missing from the Bazel build:\n%s',
            plural(missing, 'file', are=True),
            '\n'.join(str(x) for x in missing),
        )

    return missing


def check_gn_build_for_files(
    gn_extensions_to_check: Container[str],
    files: Iterable[Path],
    gn_dirs: Iterable[tuple[Path, Path]] = (),
    gn_build_files: Iterable[Path] = (),
) -> list[Path]:
    """Checks that source files are in the GN build.

    Args:
        gn_extensions_to_check: which file suffixes to look for in GN
        files: the files that should be checked
        gn_dirs: (source_dir, output_dir) tuples with which to run gn desc
        gn_build_files: paths to BUILD.gn files to directly search for paths

    Returns:
        a list of missing files; will be empty if there were no missing files
    """

    # Collect all paths in GN builds.
    gn_builds: Set[Path] = set()

    for source_dir, output_dir in gn_dirs:
        gn_builds.update(
            _get_paths_from_command(source_dir, 'gn', 'desc', output_dir, '*')
        )

    gn_builds.update(_search_files_for_paths(gn_build_files))

    missing: list[Path] = []

    if gn_dirs or gn_build_files:
        for path in (p for p in files if p.suffix in gn_extensions_to_check):
            if path not in gn_builds:
                missing.append(path)

    if missing:
        _LOG.warning(
            '%s missing from the GN build:\n%s',
            plural(missing, 'file', are=True),
            '\n'.join(str(x) for x in missing),
        )

    return missing


def check_builds_for_files(
    bazel_extensions_to_check: Container[str],
    gn_extensions_to_check: Container[str],
    files: Iterable[Path],
    bazel_dirs: Iterable[Path] = (),
    gn_dirs: Iterable[tuple[Path, Path]] = (),
    gn_build_files: Iterable[Path] = (),
) -> dict[str, list[Path]]:
    """Checks that source files are in the GN and Bazel builds.

    Args:
        bazel_extensions_to_check: which file suffixes to look for in Bazel
        gn_extensions_to_check: which file suffixes to look for in GN
        files: the files that should be checked
        bazel_dirs: directories in which to run bazel query
        gn_dirs: (source_dir, output_dir) tuples with which to run gn desc
        gn_build_files: paths to BUILD.gn files to directly search for paths

    Returns:
        a dictionary mapping build system ('Bazel' or 'GN' to a list of missing
        files; will be empty if there were no missing files
    """

    bazel_missing = check_bazel_build_for_files(
        bazel_extensions_to_check=bazel_extensions_to_check,
        files=files,
        bazel_dirs=bazel_dirs,
    )
    gn_missing = check_gn_build_for_files(
        gn_extensions_to_check=gn_extensions_to_check,
        files=files,
        gn_dirs=gn_dirs,
        gn_build_files=gn_build_files,
    )

    result = {}
    if bazel_missing:
        result['Bazel'] = bazel_missing
    if gn_missing:
        result['GN'] = gn_missing
    return result


@contextlib.contextmanager
def test_server(executable: str, output_dir: Path):
    """Context manager that runs a test server executable.

    Args:
        executable: name of the test server executable
        output_dir: path to the output directory (for logs)
    """

    with open(output_dir / 'test_server.log', 'w') as outs:
        try:
            proc = subprocess.Popen(
                [executable, '--verbose'],
                stdout=outs,
                stderr=subprocess.STDOUT,
            )

            yield

        finally:
            proc.terminate()  # pylint: disable=used-before-assignment


@contextlib.contextmanager
def modified_env(**envvars):
    """Context manager that sets environment variables.

    Use by assigning values to variable names in the argument list, e.g.:
        `modified_env(MY_FLAG="some value")`

    Args:
        envvars: Keyword arguments
    """
    saved_env = os.environ.copy()
    os.environ.update(envvars)
    try:
        yield
    finally:
        os.environ = saved_env


def fuzztest_prng_seed(ctx: PresubmitContext) -> str:
    """Convert the RNG seed to the format expected by FuzzTest.

    FuzzTest can be configured to use the seed by setting the
    `FUZZTEST_PRNG_SEED` environment variable to this value.

    Args:
        ctx: The context that includes a pseudorandom number generator seed.
    """
    rng_bytes = ctx.rng_seed.to_bytes(32, sys.byteorder)
    return base64.urlsafe_b64encode(rng_bytes).decode('ascii').rstrip('=')


@filter_paths(
    file_filter=FileFilter(
        endswith=('.bzl', '.bazel'),
        name=('WORKSPACE',),
        exclude=(r'pw_presubmit/py/pw_presubmit/format/test_data',),
    )
)
def bazel_lint(ctx: PresubmitContext):
    """Runs buildifier with lint on Bazel files.

    Should be run after bazel_format since that will give more useful output
    for formatting-only issues.
    """

    failure = False
    for path in ctx.paths:
        try:
            call('buildifier', '--lint=warn', '--mode=check', path)
        except PresubmitFailure:
            failure = True

    if failure:
        raise PresubmitFailure


@Check
def gn_gen_check(ctx: PresubmitContext):
    """Runs gn gen --check to enforce correct header dependencies."""
    gn_gen(ctx, gn_check=True)


Item = int | str
Value = Item | Sequence[Item]
ValueCallable = Callable[[PresubmitContext], Value]
InputItem = Item | ValueCallable
InputValue = InputItem | Sequence[InputItem]


def _value(ctx: PresubmitContext, val: InputValue) -> Value:
    """Process any lambdas inside val

    val is a single value or a list of values, any of which might be a lambda
    that needs to be resolved. Call each of these lambdas with ctx and replace
    the lambda with the result. Return the updated top-level structure.
    """

    if isinstance(val, (str, int)):
        return val
    if callable(val):
        return val(ctx)

    result: list[Item] = []
    for item in val:
        if callable(item):
            call_result = item(ctx)
            if isinstance(call_result, (int, str)):
                result.append(call_result)
            else:  # Sequence.
                result.extend(call_result)
        elif isinstance(item, (int, str)):
            result.append(item)
        else:  # Sequence.
            result.extend(item)
    return result


_CtxMgrLambda = Callable[[PresubmitContext], ContextManager]
_CtxMgrOrLambda = ContextManager | _CtxMgrLambda


@dataclass(frozen=True)
class CommonCoverageOptions:
    """Coverage options shared by both CodeSearch and Gerrit.

    For Google use only.
    """

    # The "root" of the Kalypsi GCS bucket path to which the coverage data
    # should be uploaded. Typically gs://ng3-metrics/ng3-<teamname>-coverage.
    target_bucket_root: str

    # The project name in the Kalypsi GCS bucket path.
    target_bucket_project: str

    # See go/kalypsi-abs#trace-type-required.
    trace_type: str

    # go/kalypsi-abs#owner-required.
    owner: str

    # go/kalypsi-abs#bug-component-required.
    bug_component: str


@dataclass(frozen=True)
class CodeSearchCoverageOptions:
    """CodeSearch-specific coverage options. For Google use only."""

    # The name of the Gerrit host containing the CodeSearch repo. Just the name
    # ("pigweed"), not the full URL ("pigweed.googlesource.com"). This may be
    # different from the host from which the code was originally checked out.
    host: str

    # The name of the project, as expected by CodeSearch. Typically
    # 'codesearch'.
    project: str

    # See go/kalypsi-abs#ref-required.
    ref: str

    # See go/kalypsi-abs#source-required.
    source: str

    # See go/kalypsi-abs#add-prefix-optional.
    add_prefix: str = ''


@dataclass(frozen=True)
class GerritCoverageOptions:
    """Gerrit-specific coverage options. For Google use only."""

    # The name of the project, as expected by Gerrit. This is typically the
    # repository name, e.g. 'pigweed/pigweed' for upstream Pigweed.
    # See go/kalypsi-inc#project-required.
    project: str


@dataclass(frozen=True)
class CoverageOptions:
    """Coverage collection configuration. For Google use only."""

    common: CommonCoverageOptions
    codesearch: tuple[CodeSearchCoverageOptions, ...]
    gerrit: GerritCoverageOptions


class _NinjaBase(Check):
    """Thin wrapper of Check for steps that call ninja."""

    def __init__(
        self,
        *args,
        packages: Sequence[str] = (),
        ninja_contexts: Sequence[_CtxMgrOrLambda] = (),
        ninja_targets: str | Sequence[str] | Sequence[Sequence[str]] = (),
        coverage_options: CoverageOptions | None = None,
        **kwargs,
    ):
        """Initializes a _NinjaBase object.

        Args:
            *args: Passed on to superclass.
            packages: List of 'pw package' packages to install.
            ninja_contexts: List of context managers to apply around ninja
                calls.
            ninja_targets: Single ninja target, list of Ninja targets, or list
                of list of ninja targets. If a list of a list, ninja will be
                called multiple times with the same build directory.
            coverage_options: Coverage collection options (or None, if not
                collecting coverage data).
            **kwargs: Passed on to superclass.
        """
        super().__init__(*args, **kwargs)
        self._packages: Sequence[str] = packages
        self._ninja_contexts: tuple[_CtxMgrOrLambda, ...] = tuple(
            ninja_contexts
        )
        self._coverage_options = coverage_options

        if isinstance(ninja_targets, str):
            ninja_targets = (ninja_targets,)
        ninja_targets = list(ninja_targets)
        all_strings = all(isinstance(x, str) for x in ninja_targets)
        any_strings = any(isinstance(x, str) for x in ninja_targets)
        if ninja_targets and all_strings != any_strings:
            raise ValueError(repr(ninja_targets))

        self._ninja_target_lists: tuple[tuple[str, ...], ...]
        if all_strings:
            targets: list[str] = []
            for target in ninja_targets:
                targets.append(target)  # type: ignore
            self._ninja_target_lists = (tuple(targets),)
        else:
            self._ninja_target_lists = tuple(tuple(x) for x in ninja_targets)

    @property
    def ninja_targets(self) -> list[str]:
        return list(itertools.chain(*self._ninja_target_lists))

    def _install_package(  # pylint: disable=no-self-use
        self,
        ctx: PresubmitContext,
        package: str,
    ) -> PresubmitResult:
        install_package(ctx, package)
        return PresubmitResult.PASS

    @contextlib.contextmanager
    def _context(self, ctx: PresubmitContext):
        """Apply any context managers necessary for building."""
        with contextlib.ExitStack() as stack:
            for mgr in self._ninja_contexts:
                if isinstance(mgr, contextlib.AbstractContextManager):
                    stack.enter_context(mgr)
                else:
                    stack.enter_context(mgr(ctx))  # type: ignore
            yield

    def _ninja(
        self, ctx: PresubmitContext, targets: Sequence[str]
    ) -> PresubmitResult:
        with self._context(ctx):
            ninja(ctx, *targets)
        return PresubmitResult.PASS

    def _coverage(
        self, ctx: PresubmitContext, options: CoverageOptions
    ) -> PresubmitResult:
        """Archive and (on LUCI) upload coverage reports."""
        reports = ctx.output_dir / 'coverage_reports'
        os.makedirs(reports, exist_ok=True)
        coverage_jsons: list[Path] = []
        for path in ctx.output_dir.rglob('coverage_report'):
            _LOG.debug('exploring %s', path)
            name = str(path.relative_to(ctx.output_dir))
            name = name.replace('_', '').replace('/', '_')
            with tarfile.open(reports / f'{name}.tar.gz', 'w:gz') as tar:
                tar.add(path, arcname=name, recursive=True)
            json_path = path / 'json' / 'report.json'
            if json_path.is_file():
                _LOG.debug('found json %s', json_path)
                coverage_jsons.append(json_path)

        if not coverage_jsons:
            ctx.fail('No coverage json file found')
            return PresubmitResult.FAIL

        if len(coverage_jsons) > 1:
            _LOG.warning(
                'More than one coverage json file, selecting first: %r',
                coverage_jsons,
            )

        coverage_json = coverage_jsons[0]

        if ctx.luci:
            if not ctx.luci.is_prod:
                _LOG.warning('Not uploading coverage since not running in prod')
                return PresubmitResult.PASS

            with self._context(ctx):
                metadata_json_paths = _write_coverage_metadata(ctx, options)
                for i, metadata_json in enumerate(metadata_json_paths):
                    # GCS bucket paths are POSIX-like.
                    coverage_gcs_path = posixpath.join(
                        options.common.target_bucket_root,
                        'incremental' if ctx.luci.is_try else 'absolute',
                        options.common.target_bucket_project,
                        f'{ctx.luci.buildbucket_id}-{i}',
                    )
                    _copy_to_gcs(
                        ctx,
                        coverage_json,
                        posixpath.join(coverage_gcs_path, 'report.json'),
                    )
                    _copy_to_gcs(
                        ctx,
                        metadata_json,
                        posixpath.join(coverage_gcs_path, 'metadata.json'),
                    )

                return PresubmitResult.PASS

        _LOG.warning('Not uploading coverage since running locally')
        return PresubmitResult.PASS

    def _package_substeps(self) -> Iterator[SubStep]:
        for package in self._packages:
            yield SubStep(
                f'install {package} package',
                self._install_package,
                (package,),
            )

    def _ninja_substeps(self) -> Iterator[SubStep]:
        targets_parts = set()
        for targets in self._ninja_target_lists:
            targets_part = " ".join(targets)
            maxlen = 70
            if len(targets_part) > maxlen:
                targets_part = f'{targets_part[0:maxlen-3]}...'
            assert targets_part not in targets_parts
            targets_parts.add(targets_part)
            yield SubStep(f'ninja {targets_part}', self._ninja, (targets,))

    def _coverage_substeps(self) -> Iterator[SubStep]:
        if self._coverage_options is not None:
            yield SubStep('coverage', self._coverage, (self._coverage_options,))


def _copy_to_gcs(ctx: PresubmitContext, filepath: Path, gcs_dst: str):
    cmd = [
        "gsutil",
        "cp",
        filepath,
        gcs_dst,
    ]

    upload_stdout = ctx.output_dir / (filepath.name + '.stdout')
    with upload_stdout.open('w') as outs:
        call(*cmd, tee=outs)


def _write_coverage_metadata(
    ctx: PresubmitContext, options: CoverageOptions
) -> Sequence[Path]:
    """Write out Kalypsi coverage metadata file(s) and return their paths."""
    assert ctx.luci is not None
    assert len(ctx.luci.triggers) == 1
    change = ctx.luci.triggers[0]

    metadata = {
        'trace_type': options.common.trace_type,
        'trim_prefix': str(ctx.root),
        'patchset_num': change.patchset,
        'change_id': change.number,
        'owner': options.common.owner,
        'bug_component': options.common.bug_component,
    }

    if ctx.luci.is_try:
        # Running in CQ: uploading incremental coverage
        metadata.update(
            {
                'change_id': change.number,
                'host': change.gerrit_name,
                'patchset_num': change.patchset,
                'project': options.gerrit.project,
            }
        )

        metadata_json = ctx.output_dir / "metadata.json"
        with metadata_json.open('w') as metadata_file:
            json.dump(metadata, metadata_file)
        return (metadata_json,)

    # Running in CI: uploading absolute coverage, possibly to multiple locations
    # since a repo could be in codesearch in multiple places.
    metadata_jsons = []
    for i, cs in enumerate(options.codesearch):
        metadata.update(
            {
                'add_prefix': cs.add_prefix,
                'commit_id': change.ref,
                'host': cs.host,
                'project': cs.project,
                'ref': cs.ref,
                'source': cs.source,
            }
        )

        metadata_json = ctx.output_dir / f'metadata-{i}.json'
        with metadata_json.open('w') as metadata_file:
            json.dump(metadata, metadata_file)
        metadata_jsons.append(metadata_json)

    return tuple(metadata_jsons)


class GnGenNinja(_NinjaBase):
    """Thin wrapper of Check for steps that just call gn/ninja.

    Runs gn gen, ninja, then gn check.
    """

    def __init__(
        self,
        *args,
        gn_args: (  # pylint: disable=redefined-outer-name
            dict[str, Any] | None
        ) = None,
        **kwargs,
    ):
        """Initializes a GnGenNinja object.

        Args:
            *args: Passed on to superclass.
            gn_args: dict of GN args.
            **kwargs: Passed on to superclass.
        """
        super().__init__(self._substeps(), *args, **kwargs)
        self._gn_args: dict[str, Any] = gn_args or {}

    def add_default_gn_args(self, args):
        """Add any project-specific default GN args to 'args'."""

    @property
    def gn_args(self) -> dict[str, Any]:
        return self._gn_args

    def _gn_gen(self, ctx: PresubmitContext) -> PresubmitResult:
        args: dict[str, Any] = {}
        if self._coverage_options is not None:
            args['pw_toolchain_COVERAGE_ENABLED'] = True
            args['pw_build_PYTHON_TEST_COVERAGE'] = True

            if ctx.incremental:
                args['pw_toolchain_PROFILE_SOURCE_FILES'] = [
                    f'//{x.relative_to(ctx.root)}' for x in ctx.paths
                ]

        self.add_default_gn_args(args)

        args.update({k: _value(ctx, v) for k, v in self._gn_args.items()})
        gn_gen(ctx, gn_check=False, **args)  # type: ignore
        return PresubmitResult.PASS

    def _substeps(self) -> Iterator[SubStep]:
        yield from self._package_substeps()

        yield SubStep('gn gen', self._gn_gen)

        yield from self._ninja_substeps()

        # Run gn check after building so it can check generated files.
        yield SubStep('gn check', gn_check)

        yield from self._coverage_substeps()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_thread_protos/thread_snapshot_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_thread_protos import thread_pb2 as pw__thread__protos_dot_thread__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.pw_thread_protos/thread_snapshot_service.proto\x12\x0fpw.thread.proto\x1a\x1dpw_thread_protos/thread.proto\"+\n\rThreadRequest\x12\x11\n\x04name\x18\x01 \x01(\x0cH\x00\x88\x01\x01\x42\x07\n\x05_name2u\n\x15ThreadSnapshotService\x12\\\n\x11GetPeakStackUsage\x12\x1e.pw.thread.proto.ThreadRequest\x1a#.pw.thread.proto.SnapshotThreadInfo\"\x00\x30\x01\x42\x19\n\x0fpw.thread.protoB\x06Threadb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_thread_protos.thread_snapshot_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\017pw.thread.protoB\006Thread'
  _globals['_THREADREQUEST']._serialized_start=98
  _globals['_THREADREQUEST']._serialized_end=141
  _globals['_THREADSNAPSHOTSERVICE']._serialized_start=143
  _globals['_THREADSNAPSHOTSERVICE']._serialized_end=260
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2023 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _DigitalIoState:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _DigitalIoStateEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_DigitalIoState.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    INACTIVE: _DigitalIoState.ValueType  # 0
    ACTIVE: _DigitalIoState.ValueType  # 1

class DigitalIoState(_DigitalIoState, metaclass=_DigitalIoStateEnumTypeWrapper): ...

INACTIVE: DigitalIoState.ValueType  # 0
ACTIVE: DigitalIoState.ValueType  # 1
global___DigitalIoState = DigitalIoState

@typing_extensions.final
class DigitalIoEnableRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LINE_INDEX_FIELD_NUMBER: builtins.int
    ENABLE_FIELD_NUMBER: builtins.int
    line_index: builtins.int
    """Which line to select."""
    enable: builtins.bool
    """Whether to enable or disable."""
    def __init__(
        self,
        *,
        line_index: builtins.int = ...,
        enable: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["enable", b"enable", "line_index", b"line_index"]) -> None: ...

global___DigitalIoEnableRequest = DigitalIoEnableRequest

@typing_extensions.final
class DigitalIoEnableResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___DigitalIoEnableResponse = DigitalIoEnableResponse

@typing_extensions.final
class DigitalIoSetStateRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LINE_INDEX_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    line_index: builtins.int
    """Which line to select."""
    state: global___DigitalIoState.ValueType
    """Which state to set to."""
    def __init__(
        self,
        *,
        line_index: builtins.int = ...,
        state: global___DigitalIoState.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["line_index", b"line_index", "state", b"state"]) -> None: ...

global___DigitalIoSetStateRequest = DigitalIoSetStateRequest

@typing_extensions.final
class DigitalIoSetStateResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___DigitalIoSetStateResponse = DigitalIoSetStateResponse

@typing_extensions.final
class DigitalIoGetStateRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LINE_INDEX_FIELD_NUMBER: builtins.int
    line_index: builtins.int
    """Which line to select."""
    def __init__(
        self,
        *,
        line_index: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["line_index", b"line_index"]) -> None: ...

global___DigitalIoGetStateRequest = DigitalIoGetStateRequest

@typing_extensions.final
class DigitalIoGetStateResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATE_FIELD_NUMBER: builtins.int
    state: global___DigitalIoState.ValueType
    def __init__(
        self,
        *,
        state: global___DigitalIoState.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["state", b"state"]) -> None: ...

global___DigitalIoGetStateResponse = DigitalIoGetStateResponse



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""Utils to decode logs."""

from dataclasses import dataclass
import datetime
import logging
import re
from typing import Any, Callable

from pw_log.proto import log_pb2
import pw_log_tokenized
import pw_status
from pw_tokenizer import Detokenizer
from pw_tokenizer.proto import decode_optionally_tokenized

_LOG = logging.getLogger(__name__)


@dataclass(frozen=True)
class LogLineLevel:
    """Tuple of line number and level packed in LogEntry."""

    line: int
    level: int


class Log:
    """A decoded, human-readable representation of a LogEntry.

    Contains fields to represent a decoded pw_log/log.proto LogEntry message in
    a human readable way.

    Attributes:
        message: The log message as a string.
        level: A integer representing the log level, follows logging levels.
        flags: An integer with the bit flags.
        timestamp: A string representation of a timestamp.
        module_name: The module name as a string.
        thread_name: The thread name as a string.
        source_name: The source name as a string.
        file_and_line: The filepath and line as a string.
        metadata_fields: Extra fields with string-string mapping.
    """

    _LOG_LEVEL_NAMES = {
        logging.DEBUG: 'DBG',
        logging.INFO: 'INF',
        logging.WARNING: 'WRN',
        logging.ERROR: 'ERR',
        logging.CRITICAL: 'CRT',
        # The logging module does not have a FATAL level. This module's log
        # level values are 10 * PW_LOG_LEVEL values. PW_LOG_LEVEL_FATAL is 7.
        70: 'FTL',
    }
    _LEVEL_MASK = 0x7  # pylint: disable=C0103
    _LINE_OFFSET = 3  # pylint: disable=C0103

    def __init__(
        self,
        message: str = '',
        level: int = logging.NOTSET,
        flags: int = 0,
        timestamp: str = '',
        module_name: str = '',
        thread_name: str = '',
        source_name: str = '',
        file_and_line: str = '',
        metadata_fields: dict[str, str] | None = None,
    ) -> None:
        self.message = message
        self.level = level  # Value from logging levels.
        self.flags = flags
        self.timestamp = timestamp  # A human readable value.
        self.module_name = module_name
        self.thread_name = thread_name
        self.source_name = source_name
        self.file_and_line = file_and_line
        self.metadata_fields = dict()
        if metadata_fields:
            self.metadata_fields.update(metadata_fields)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Log):
            return False

        return (
            self.message == other.message
            and self.level == other.level
            and self.flags == other.flags
            and self.timestamp == other.timestamp
            and self.module_name == other.module_name
            and self.thread_name == other.thread_name
            and self.source_name == other.source_name
            and self.file_and_line == other.file_and_line
            and self.metadata_fields == other.metadata_fields
        )

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        level_name = self._LOG_LEVEL_NAMES.get(self.level, '')
        metadata = ' '.join(map(str, self.metadata_fields.values()))
        return (
            f'{level_name} [{self.source_name}] {self.module_name} '
            f'{self.timestamp} {self.message} {self.file_and_line} '
            f'{metadata}'
        ).strip()

    @staticmethod
    def pack_line_level(line: int, logging_log_level: int) -> int:
        """Packs the line and level values into an integer as done in LogEntry.

        Args:
            line: the line number
            level: the logging level using logging levels.
        Returns:
            An integer with the two values bitpacked.
        """
        return (line << Log._LINE_OFFSET) | (
            Log.logging_level_to_pw_log_level(logging_log_level)
            & Log._LEVEL_MASK
        )

    @staticmethod
    def unpack_line_level(packed_line_level: int) -> LogLineLevel:
        """Unpacks the line and level values packed as done in LogEntry.

        Args:
            An integer with the two values bitpacked.
        Returns:
            line: the line number
            level: the logging level using logging levels.
        """
        line = packed_line_level >> Log._LINE_OFFSET
        # Convert to logging module level number.
        level = Log.pw_log_level_to_logging_level(
            packed_line_level & Log._LEVEL_MASK
        )
        return LogLineLevel(line=line, level=level)

    @staticmethod
    def pw_log_level_to_logging_level(pw_log_level: int) -> int:
        """Maps a pw_log/levels.h value to Python logging level value."""
        return pw_log_level * 10

    @staticmethod
    def logging_level_to_pw_log_level(logging_log_level: int) -> int:
        """Maps a Python logging level value to a pw_log/levels.h value."""
        return int(logging_log_level / 10)


def pw_status_code_to_name(
    message: str, status_pattern: str = 'pw::Status=([0-9]+)'
) -> str:
    """Replaces the pw::Status code number with the status name.

    Searches for a matching pattern encapsulating the status code number and
    replaces it with the status name. This is useful when logging the status
    code instead of the string to save space.

    Args:
        message: The string to look for the status code.
        status_pattern: The regex pattern describing the format encapsulating
          the status code.
    Returns:
       The message string with the status name if found, else the original
       string.
    """
    # Min and max values for Status.
    max_status_value = max(status.value for status in pw_status.Status)
    min_status_value = min(status.value for status in pw_status.Status)

    def replacement_callback(match: re.Match) -> str:
        status_code = int(match.group(1))
        if min_status_value <= status_code <= max_status_value:
            return pw_status.Status(status_code).name
        return match.group(0)

    return re.sub(
        pattern=status_pattern, repl=replacement_callback, string=message
    )


def log_decoded_log(
    log: Log, logger: logging.Logger | logging.LoggerAdapter
) -> None:
    """Formats and saves the log information in a pw console compatible way.

    Arg:
        logger: The logger to emit the log information to.
    """
    # Fields used for pw console table view.
    log.metadata_fields['module'] = log.module_name
    log.metadata_fields['source_name'] = log.source_name
    log.metadata_fields['timestamp'] = log.timestamp
    log.metadata_fields['msg'] = log.message
    log.metadata_fields['file'] = log.file_and_line

    logger.log(
        log.level,
        '[%s] %s %s %s %s',
        log.source_name,
        log.module_name,
        log.timestamp,
        log.message,
        log.file_and_line,
        extra=dict(extra_metadata_fields=log.metadata_fields),
    )


def _timestamp_format(
    timestamp: int,
    input_resolution: int,
    truncate_microseconds_to_three_digits: bool = False,
) -> str:
    time_delta = datetime.timedelta(seconds=timestamp / input_resolution)

    minutes, seconds = divmod(time_delta.seconds, 60)
    hours, minutes = divmod(minutes, 60)
    microseconds = time_delta.microseconds

    hms_str = f'{hours:02}:{minutes:02}:{seconds:02}'
    microseconds_str = f'{microseconds:06}'
    if truncate_microseconds_to_three_digits:
        microseconds_str = microseconds_str[:-3]

    return f'{hms_str}.{microseconds_str}'


def timestamp_parser_ms_since_boot(timestamp: int) -> str:
    """Decodes timestamp as milliseconds since boot.

    Args:
        timestamp: The timestamp as an integer.
    Returns:
        A string representation of the timestamp.
    """
    return _timestamp_format(
        timestamp,
        input_resolution=10**3,
        truncate_microseconds_to_three_digits=True,
    )


def timestamp_parser_ns_since_boot(timestamp: int) -> str:
    """Decodes timestamp as nanoseconds since boot.

    Args:
        timestamp: The timestamp as an integer.
    Returns:
        A string representation of the timestamp.
    """
    return _timestamp_format(timestamp, input_resolution=10**9)


class LogStreamDecoder:
    """Decodes an RPC stream of LogEntries packets.

    Performs log drop detection on the stream of LogEntries proto messages.

    Args:
        decoded_log_handler: Callback called on each decoded log.
        detokenizer: Detokenizes log messages if tokenized when provided.
        source_name: Optional string to identify the logs source.
        timestamp_parser: Optional timestamp parser number to a string.
        message_parser: Optional message parser called after detokenization is
          attempted on a log message.
    """

    DROP_REASON_LOSS_AT_TRANSPORT = 'loss at transport'
    DROP_REASON_SOURCE_NOT_CONNECTED = 'source not connected'
    DROP_REASON_SOURCE_ENQUEUE_FAILURE = 'enqueue failure at source'

    def __init__(
        self,
        decoded_log_handler: Callable[[Log], None],
        detokenizer: Detokenizer | None = None,
        source_name: str = '',
        timestamp_parser: Callable[[int], str] | None = None,
        message_parser: Callable[[str], str] | None = None,
    ):
        self.decoded_log_handler = decoded_log_handler
        self.detokenizer = detokenizer
        self.source_name = source_name
        self.timestamp_parser = timestamp_parser
        self.message_parser = message_parser
        self._expected_log_sequence_id = 0

    def parse_log_entries_proto(
        self, log_entries_proto: log_pb2.LogEntries
    ) -> None:
        """Parses each LogEntry in log_entries_proto.

        Args:
            log_entry_proto: A LogEntry message proto.
        Returns:
            A Log object with the decoded log_entry_proto.
        """
        has_received_logs = self._expected_log_sequence_id > 0
        dropped_log_count = self._calculate_dropped_logs(log_entries_proto)
        if dropped_log_count > 0:
            reason = (
                self.DROP_REASON_LOSS_AT_TRANSPORT
                if has_received_logs
                else self.DROP_REASON_SOURCE_NOT_CONNECTED
            )
            self.decoded_log_handler(
                self._handle_log_drop_count(dropped_log_count, reason)
            )
        elif dropped_log_count < 0:
            _LOG.error('Log sequence ID is smaller than expected')

        for i, log_entry_proto in enumerate(log_entries_proto.entries):
            # Handle dropped count first.
            if log_entry_proto.dropped:
                # Avoid duplicating drop reports since the device will report
                # a drop count due to a transmission failure, of the last
                # attempted transmission only, in the first entry of the next
                # successful transmission.
                if i == 0 and dropped_log_count >= log_entry_proto.dropped:
                    continue
            parsed_log = self.parse_log_entry_proto(log_entry_proto)
            self.decoded_log_handler(parsed_log)

    def parse_log_entry_proto(self, log_entry_proto: log_pb2.LogEntry) -> Log:
        """Parses the log_entry_proto contents into a human readable format.

        Args:
            log_entry_proto: A LogEntry message proto.
        Returns:
            A Log object with the decoded log_entry_proto.
        """
        detokenized_message = decode_optionally_tokenized(
            self.detokenizer, log_entry_proto.message
        )
        # Handle dropped count first.
        if log_entry_proto.dropped:
            drop_reason = self.DROP_REASON_SOURCE_ENQUEUE_FAILURE
            if detokenized_message:
                drop_reason = detokenized_message.lower()
            return self._handle_log_drop_count(
                log_entry_proto.dropped, drop_reason
            )

        # Parse message and metadata, if any, encoded in a key-value format as
        # described in pw_log/log.proto LogEntry::message field.
        message_and_metadata = pw_log_tokenized.FormatStringWithMetadata(
            detokenized_message
        )
        module_name = decode_optionally_tokenized(
            self.detokenizer, log_entry_proto.module
        )
        if not module_name:
            module_name = message_and_metadata.module

        line_level_tuple = Log.unpack_line_level(log_entry_proto.line_level)
        file_and_line = decode_optionally_tokenized(
            self.detokenizer, log_entry_proto.file
        )
        if not file_and_line:
            # Set file name if found in the metadata.
            if message_and_metadata.file is not None:
                file_and_line = message_and_metadata.file
            else:
                file_and_line = ''

        # Add line number to filepath if needed.
        if line_level_tuple.line and ':' not in file_and_line:
            file_and_line += f':{line_level_tuple.line}'
        # Add extra log information avoiding duplicated data.
        metadata_fields = {
            k: v
            for k, v in message_and_metadata.fields.items()
            if k not in ['file', 'module', 'msg']
        }

        message = message_and_metadata.message
        if self.message_parser:
            message = self.message_parser(message)
        if self.timestamp_parser:
            timestamp = self.timestamp_parser(log_entry_proto.timestamp)
        else:
            timestamp = str(log_entry_proto.timestamp)
        log = Log(
            message=message,
            level=line_level_tuple.level,
            flags=log_entry_proto.flags,
            timestamp=timestamp,
            module_name=module_name,
            thread_name=decode_optionally_tokenized(
                self.detokenizer, log_entry_proto.thread
            ),
            source_name=self.source_name,
            file_and_line=file_and_line,
            metadata_fields=metadata_fields,
        )

        return log

    def _handle_log_drop_count(self, drop_count: int, reason: str) -> Log:
        log_word = 'logs' if drop_count > 1 else 'log'
        log = Log(
            message=f'Dropped {drop_count} {log_word} due to {reason}',
            level=logging.WARNING,
            source_name=self.source_name,
        )
        return log

    def _calculate_dropped_logs(
        self, log_entries_proto: log_pb2.LogEntries
    ) -> int:
        # Count log messages received that don't use the dropped field.
        messages_received = sum(
            1 if not log_proto.dropped else 0
            for log_proto in log_entries_proto.entries
        )
        dropped_log_count = (
            log_entries_proto.first_entry_sequence_id
            - self._expected_log_sequence_id
        )
        self._expected_log_sequence_id = (
            log_entries_proto.first_entry_sequence_id + messages_received
        )
        return dropped_log_count



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Wrapers for socket clients to log read and write data."""
from __future__ import annotations

from typing import Callable, TYPE_CHECKING

import errno
import re
import socket

from pw_console.plugins.bandwidth_toolbar import SerialBandwidthTracker

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer


class SocketClient:
    """Socket transport implementation."""

    FILE_SOCKET_SERVER = 'file'
    DEFAULT_SOCKET_SERVER = 'localhost'
    DEFAULT_SOCKET_PORT = 33000
    PW_RPC_MAX_PACKET_SIZE = 256
    DEFAULT_TIMEOUT = 0.5

    _InitArgsType = tuple[
        socket.AddressFamily, int  # pylint: disable=no-member
    ]
    # Can be a string, (address, port) for AF_INET or (address, port, flowinfo,
    # scope_id) AF_INET6.
    _AddressType = str | tuple[str, int] | tuple[str, int, int, int]

    def __init__(
        self,
        config: str,
        on_disconnect: Callable[[SocketClient], None] | None = None,
        timeout: float | None = None,
    ):
        """Creates a socket connection.

        Args:
          config: The socket configuration. Accepted values and formats are:
            'default' - uses the default configuration (localhost:33000)
            'address:port' - An IPv4 address and port.
            'address' - An IPv4 address. Uses default port 33000.
            '[address]:port' - An IPv6 address and port.
            '[address]' - An IPv6 address. Uses default port 33000.
            'file:path_to_file' - A Unix socket at ``path_to_file``.
            In the formats above,``address`` can be an actual address or a name
            that resolves to an address through name-resolution.
          on_disconnect: An optional callback called when the socket
            disconnects.

        Raises:
          TypeError: The type of socket is not supported.
          ValueError: The socket configuration is invalid.
        """
        self.socket: socket.socket
        (
            self._socket_init_args,
            self._address,
        ) = SocketClient._parse_socket_config(config)
        self._on_disconnect = on_disconnect
        self._timeout = SocketClient.DEFAULT_TIMEOUT
        if timeout:
            self._timeout = timeout
        self._connected = False
        self.connect()

    @staticmethod
    def _parse_socket_config(
        config: str,
    ) -> tuple[SocketClient._InitArgsType, SocketClient._AddressType]:
        """Sets the variables used to create a socket given a config string.

        Raises:
          TypeError: The type of socket is not supported.
          ValueError: The socket configuration is invalid.
        """
        init_args: SocketClient._InitArgsType
        address: SocketClient._AddressType

        # Check if this is using the default settings.
        if config == 'default':
            init_args = socket.AF_INET6, socket.SOCK_STREAM
            address = (
                SocketClient.DEFAULT_SOCKET_SERVER,
                SocketClient.DEFAULT_SOCKET_PORT,
            )
            return init_args, address

        # Check if this is a UNIX socket.
        unix_socket_file_setting = f'{SocketClient.FILE_SOCKET_SERVER}:'
        if config.startswith(unix_socket_file_setting):
            # Unix socket support is available on Windows 10 since April
            # 2018. However, there is no Python support on Windows yet.
            # See https://bugs.python.org/issue33408 for more information.
            if not hasattr(socket, 'AF_UNIX'):
                raise TypeError(
                    'Unix sockets are not supported in this environment.'
                )
            init_args = (
                socket.AF_UNIX,  # pylint: disable=no-member
                socket.SOCK_STREAM,
            )
            address = config[len(unix_socket_file_setting) :]
            return init_args, address

        # Search for IPv4 or IPv6 address or name and port.
        # First, try to capture an IPv6 address as anything inside []. If there
        # are no [] capture the IPv4 address. Lastly, capture the port as the
        # numbers after :, if any.
        match = re.match(
            r'(\[(?P<ipv6_addr>.+)\]:?|(?P<ipv4_addr>[a-zA-Z0-9\._\/]+):?)'
            r'(?P<port>[0-9]+)?',
            config,
        )
        invalid_config_message = (
            f'Invalid socket configuration "{config}"'
            'Accepted values are "default", "file:<file_path>", '
            '"<name_or_ipv4_address>" with optional ":<port>", and '
            '"[<name_or_ipv6_address>]" with optional ":<port>".'
        )
        if match is None:
            raise ValueError(invalid_config_message)

        info = match.groupdict()
        if info['port']:
            port = int(info['port'])
        else:
            port = SocketClient.DEFAULT_SOCKET_PORT

        if info['ipv4_addr']:
            ip_addr = info['ipv4_addr']
        elif info['ipv6_addr']:
            ip_addr = info['ipv6_addr']
        else:
            raise ValueError(invalid_config_message)

        sock_family, sock_type, _, _, address = socket.getaddrinfo(
            ip_addr, port, type=socket.SOCK_STREAM
        )[0]
        init_args = sock_family, sock_type
        return init_args, address

    def __del__(self):
        if self._connected:
            self.socket.close()

    def write(self, data: ReadableBuffer) -> None:
        """Writes data and detects disconnects."""
        if not self._connected:
            raise Exception('Socket is not connected.')
        try:
            self.socket.sendall(data)
        except socket.error as exc:
            if isinstance(exc.args, tuple) and exc.args[0] == errno.EPIPE:
                self._handle_disconnect()
            else:
                raise exc

    def read(self, num_bytes: int = PW_RPC_MAX_PACKET_SIZE) -> bytes:
        """Blocks until data is ready and reads up to num_bytes."""
        if not self._connected:
            raise Exception('Socket is not connected.')
        data = self.socket.recv(num_bytes)
        # Since this is a blocking read, no data returned means the socket is
        # closed.
        if not data:
            self._handle_disconnect()
        return data

    def connect(self) -> None:
        """Connects to socket."""
        self.socket = socket.socket(*self._socket_init_args)

        # Enable reusing address and port for reconnections.
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if hasattr(socket, 'SO_REUSEPORT'):
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self.socket.settimeout(self._timeout)
        self.socket.connect(self._address)
        self._connected = True

    def _handle_disconnect(self):
        """Escalates a socket disconnect to the user."""
        self.socket.close()
        self._connected = False
        if self._on_disconnect:
            self._on_disconnect(self)

    def fileno(self) -> int:
        return self.socket.fileno()


class SocketClientWithLogging(SocketClient):
    """Socket with read and write wrappers for logging."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bandwidth_tracker = SerialBandwidthTracker()

    def read(
        self, num_bytes: int = SocketClient.PW_RPC_MAX_PACKET_SIZE
    ) -> bytes:
        data = super().read(num_bytes)
        self._bandwidth_tracker.track_read_data(data)
        return data

    def write(self, data: ReadableBuffer) -> None:
        self._bandwidth_tracker.track_write_data(data)
        super().write(data)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_trace_protos/trace_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_chrono_protos import chrono_pb2 as pw__chrono__protos_dot_chrono__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#pw_trace_protos/trace_service.proto\x12\x0epw.trace.proto\x1a\x1dpw_chrono_protos/chrono.proto\"\x0e\n\x0cStartRequest\"\x0f\n\rStartResponse\"\r\n\x0bStopRequest\"0\n\x0cStopResponse\x12\x14\n\x07\x66ile_id\x18\x01 \x01(\rH\x00\x88\x01\x01\x42\n\n\x08_file_id\"\x18\n\x16\x43lockParametersRequest\"O\n\x17\x43lockParametersResponse\x12\x34\n\x10\x63lock_parameters\x18\x01 \x01(\x0b\x32\x1a.pw.chrono.ClockParameters2\x84\x02\n\x0cTraceService\x12\x46\n\x05Start\x12\x1c.pw.trace.proto.StartRequest\x1a\x1d.pw.trace.proto.StartResponse\"\x00\x12\x43\n\x04Stop\x12\x1b.pw.trace.proto.StopRequest\x1a\x1c.pw.trace.proto.StopResponse\"\x00\x12g\n\x12GetClockParameters\x12&.pw.trace.proto.ClockParametersRequest\x1a\'.pw.trace.proto.ClockParametersResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_trace_protos.trace_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_STARTREQUEST']._serialized_start=86
  _globals['_STARTREQUEST']._serialized_end=100
  _globals['_STARTRESPONSE']._serialized_start=102
  _globals['_STARTRESPONSE']._serialized_end=117
  _globals['_STOPREQUEST']._serialized_start=119
  _globals['_STOPREQUEST']._serialized_end=132
  _globals['_STOPRESPONSE']._serialized_start=134
  _globals['_STOPRESPONSE']._serialized_end=182
  _globals['_CLOCKPARAMETERSREQUEST']._serialized_start=184
  _globals['_CLOCKPARAMETERSREQUEST']._serialized_end=208
  _globals['_CLOCKPARAMETERSRESPONSE']._serialized_start=210
  _globals['_CLOCKPARAMETERSRESPONSE']._serialized_end=289
  _globals['_TRACESERVICE']._serialized_start=292
  _globals['_TRACESERVICE']._serialized_end=552
# @@protoc_insertion_point(module_scope)



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Tools for managing symlinks."""

import os
from pathlib import Path


def set_symlink(target_path: Path, symlink_path: Path) -> None:
    if symlink_path.exists():
        os.remove(symlink_path)

    os.symlink(target_path, symlink_path, target_path.is_dir())



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Device classes to interact with targets via RPC."""


import logging
import os
from pathlib import Path
import tempfile
from types import ModuleType
from collections.abc import Iterable
from typing import Any, Callable, Sequence

from pw_file import file_pb2
from pw_hdlc import rpc
from pw_log import log_decoder
from pw_log_rpc import rpc_log_stream
from pw_metric import metric_parser
import pw_rpc
from pw_rpc import callback_client, console_tools, client_utils
import pw_transfer
from pw_transfer import transfer_pb2
from pw_stream import stream_readers
from pw_system import snapshot
from pw_thread import thread_analyzer
from pw_thread_protos import thread_pb2
from pw_tokenizer import detokenize
from pw_tokenizer.proto import decode_optionally_tokenized
from pw_unit_test.rpc import run_tests as pw_unit_test_run_tests, TestRecord


# Internal log for troubleshooting this tool (the console).
_LOG = logging.getLogger(__package__)

DEFAULT_DEVICE_LOGGER = logging.getLogger('rpc_device')


class Device:
    """Represents an RPC Client for a device running a Pigweed target.

    The target must have RPC support for the following services:
     - logging
     - file
     - transfer

    Note: use this class as a base for specialized device representations.
    """

    def __init__(
        # pylint: disable=too-many-arguments
        self,
        channel_id: int,
        reader: stream_readers.CancellableReader,
        write: Callable[[bytes], Any],
        proto_library: Iterable[ModuleType | Path],
        detokenizer: detokenize.Detokenizer | None = None,
        timestamp_decoder: Callable[[int], str] | None = None,
        rpc_timeout_s: float = 5,
        use_rpc_logging: bool = True,
        use_hdlc_encoding: bool = True,
        logger: logging.Logger | logging.LoggerAdapter = DEFAULT_DEVICE_LOGGER,
    ):
        self.channel_id = channel_id
        self.protos = list(proto_library)
        self.detokenizer = detokenizer
        self.rpc_timeout_s = rpc_timeout_s

        self.logger = logger
        self.logger.setLevel(logging.DEBUG)  # Allow all device logs through.

        callback_client_impl = callback_client.Impl(
            default_unary_timeout_s=self.rpc_timeout_s,
            default_stream_timeout_s=None,
        )

        def detokenize_and_log_output(data: bytes, _detokenizer=None):
            log_messages = data.decode(
                encoding='utf-8', errors='surrogateescape'
            )

            if self.detokenizer:
                log_messages = decode_optionally_tokenized(
                    self.detokenizer, data
                )

            for line in log_messages.splitlines():
                self.logger.info(line)

        # Device has a hard dependency on transfer_pb2, so ensure it's
        # always been added to the list of compiled protos, rather than
        # requiring all clients to include it.
        if transfer_pb2 not in self.protos:
            self.protos.append(transfer_pb2)

        self.client: client_utils.RpcClient
        if use_hdlc_encoding:
            channels = [
                pw_rpc.Channel(self.channel_id, rpc.channel_output(write))
            ]
            self.client = rpc.HdlcRpcClient(
                reader,
                self.protos,
                channels,
                detokenize_and_log_output,
                client_impl=callback_client_impl,
            )
        else:
            channel = pw_rpc.Channel(self.channel_id, write)
            self.client = client_utils.NoEncodingSingleChannelRpcClient(
                reader,
                self.protos,
                channel,
                client_impl=callback_client_impl,
            )

        if use_rpc_logging:
            # Create the log decoder used by the LogStreamHandler.

            def decoded_log_handler(log: log_decoder.Log) -> None:
                log_decoder.log_decoded_log(log, self.logger)

            self._log_decoder = log_decoder.LogStreamDecoder(
                decoded_log_handler=decoded_log_handler,
                detokenizer=self.detokenizer,
                source_name='RpcDevice',
                timestamp_parser=(
                    timestamp_decoder
                    if timestamp_decoder
                    else log_decoder.timestamp_parser_ns_since_boot
                ),
            )

            # Start listening to logs as soon as possible.
            self.log_stream_handler = rpc_log_stream.LogStreamHandler(
                self.rpcs, self._log_decoder
            )
            self.log_stream_handler.start_logging()

        # Create the transfer manager
        self.transfer_service = self.rpcs.pw.transfer.Transfer
        self.transfer_manager = pw_transfer.Manager(
            self.transfer_service,
            default_response_timeout_s=self.rpc_timeout_s,
            initial_response_timeout_s=self.rpc_timeout_s,
            default_protocol_version=pw_transfer.ProtocolVersion.LATEST,
        )

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def close(self) -> None:
        self.client.close()

    def info(self) -> console_tools.ClientInfo:
        return console_tools.ClientInfo('device', self.rpcs, self.client.client)

    @property
    def rpcs(self) -> Any:
        """Returns an object for accessing services on the specified channel."""
        return next(iter(self.client.client.channels())).rpcs

    def run_tests(self, timeout_s: float | None = 5) -> TestRecord:
        """Runs the unit tests on this device."""
        return pw_unit_test_run_tests(self.rpcs, timeout_s=timeout_s)

    def echo(self, msg: str) -> str:
        """Sends a string to the device and back, returning the result."""
        return self.rpcs.pw.rpc.EchoService.Echo(msg=msg).unwrap_or_raise().msg

    def reboot(self):
        """Triggers a reboot to run asynchronously on the device.

        This function *does not* wait for the reboot to complete."""
        # `invoke` rather than call in order to ignore the result. No result
        # will be sent when the device reboots.
        self.rpcs.pw.system.proto.DeviceService.Reboot.invoke()

    def crash(self):
        """Triggers a crash to run asynchronously on the device.

        This function *does not* wait for the crash to complete."""
        # `invoke` rather than call in order to ignore the result. No result
        # will be sent when the device crashes.
        self.rpcs.pw.system.proto.DeviceService.Crash.invoke()

    def get_and_log_metrics(self) -> dict:
        """Retrieves the parsed metrics and logs them to the console."""
        metrics = metric_parser.parse_metrics(
            self.rpcs, self.detokenizer, self.rpc_timeout_s
        )

        def print_metrics(metrics, path):
            """Traverses dictionaries, until a non-dict value is reached."""
            for path_name, metric in metrics.items():
                if isinstance(metric, dict):
                    print_metrics(metric, path + '/' + path_name)
                else:
                    _LOG.info('%s/%s: %s', path, path_name, str(metric))

        print_metrics(metrics, '')
        return metrics

    def snapshot_peak_stack_usage(self, thread_name: str | None = None):
        snapshot_service = self.rpcs.pw.thread.proto.ThreadSnapshotService
        _, rsp = snapshot_service.GetPeakStackUsage(name=thread_name)

        thread_info = thread_pb2.SnapshotThreadInfo()
        for thread_info_block in rsp:
            for thread in thread_info_block.threads:
                thread_info.threads.append(thread)
        for line in str(
            thread_analyzer.ThreadSnapshotAnalyzer(thread_info)
        ).splitlines():
            _LOG.info('%s', line)

    def list_files(self) -> Sequence[file_pb2.ListResponse]:
        """Lists all files on this device.
        Returns:
            A sequence of responses from the List() RPC.
        """
        fs_service = self.rpcs.pw.file.FileSystem
        stream_response = fs_service.List()
        if not stream_response.status.ok():
            _LOG.error('Failed to list files %s', stream_response.status)
            return []

        return stream_response.responses

    def delete_file(self, path: str) -> bool:
        """Delete a file on this device.
        Args:
            path: The path of the file to delete.
        Returns:
            True on successful deletion, False on failure.
        """

        fs_service = self.rpcs.pw.file.FileSystem
        req = file_pb2.DeleteRequest(path=path)
        stream_response = fs_service.Delete(req)
        if not stream_response.status.ok():
            _LOG.error(
                'Failed to delete file %s file: %s',
                path,
                stream_response.status,
            )
            return False

        return True

    def transfer_file(self, file_id: int, dest_path: str) -> bool:
        """Transfer a file on this device to the host.
        Args:
            file_id: The file_id of the file to transfer from device.
            dest_path: The destination path to save the file to on the host.
        Returns:
            True on successful transfer, False on failure.
        Raises:
            pw_transfer.Error the transfer failed.
        """
        try:
            data = self.transfer_manager.read(file_id)
            with open(dest_path, "wb") as bin_file:
                bin_file.write(data)
            _LOG.info(
                'Successfully wrote file to %s', os.path.abspath(dest_path)
            )
        except pw_transfer.Error:
            _LOG.exception('Failed to transfer file_id %i', file_id)
            return False

        return True

    def get_crash_snapshots(self, crash_log_path: str | None = None) -> bool:
        r"""Transfer any crash snapshots on this device to the host.
        Args:
            crash_log_path: The host path to store the crash files.
              If not specified, defaults to `/tmp` or `C:\TEMP` on Windows.
        Returns:
            True on successful download of snapshot, or no snapshots
            on device.  False on failure to download snapshot.
        """
        if crash_log_path is None:
            crash_log_path = tempfile.gettempdir()

        snapshot_paths: list[file_pb2.Path] = []
        for response in self.list_files():
            for snapshot_path in response.paths:
                if snapshot_path.path.startswith('/snapshots/crash_'):
                    snapshot_paths.append(snapshot_path)

        if len(snapshot_paths) == 0:
            _LOG.info('No crash snapshot on the device.')
            return True

        for snapshot_path in snapshot_paths:
            dest_snapshot = os.path.join(
                crash_log_path, os.path.basename(snapshot_path.path)
            )
            if not self.transfer_file(snapshot_path.file_id, dest_snapshot):
                return False

            decoded_snapshot: str
            with open(dest_snapshot, 'rb') as f:
                decoded_snapshot = snapshot.decode_snapshot(
                    self.detokenizer, f.read()
                )

            dest_text_snapshot = dest_snapshot.replace(".snapshot", ".txt")
            with open(dest_text_snapshot, 'w') as f:
                f.write(decoded_snapshot)
            _LOG.info('Wrote crash snapshot to: %s', dest_text_snapshot)

            if not self.delete_file(snapshot_path.path):
                return False

        return True



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and check status of Zephyr."""
import importlib.resources
import json
import pathlib
import subprocess
import sys
import tempfile

from typing import Sequence

import pw_env_setup.virtualenv_setup

import pw_package.git_repo
import pw_package.package_manager

# Main branch, this commit is close to the v3.6 RC3 tag which contains some
# bug fixes for Twister and support for GTEST_SKIP()
_ZEPHYR_COMMIT_SHA = 'f9778472105d756fff7d1e5b54353421d356ed43'


class Zephyr(pw_package.git_repo.GitRepo):
    """Install and check status of Zephyr."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            name='zephyr',
            url=(
                'https://pigweed.googlesource.com/third_party/'
                'github/zephyrproject-rtos/zephyr'
            ),
            commit=_ZEPHYR_COMMIT_SHA,
            **kwargs,
        )

    def info(self, path: pathlib.Path) -> Sequence[str]:
        return (
            f'{self.name} installed in: {path}',
            'Enable by running "gn args out" and adding this line:',
            f'  dir_pw_third_party_zephyr = "{path}"',
        )

    @staticmethod
    def __populate_download_cache_from_cipd(path: pathlib.Path) -> None:
        """Check for Zephyr SDK in cipd"""
        package_path = path.parent.resolve()
        core_cache_path = package_path / 'zephyr_sdk'
        core_cache_path.mkdir(parents=True, exist_ok=True)

        cipd_package_subpath = 'infra/3pp/tools/zephyr_sdk/${platform}'

        # Check if the zephyr_sdk cipd package is readable
        with tempfile.NamedTemporaryFile(
            prefix='cipd', delete=True
        ) as temp_json:
            temp_json_path = pathlib.Path(temp_json.name)
            cipd_acl_check_command = [
                'cipd',
                'acl-check',
                cipd_package_subpath,
                '-reader',
                '-json-output',
                str(temp_json_path),
            ]
            subprocess.run(cipd_acl_check_command, capture_output=True)

            # Return if no packages are readable.
            if not temp_json_path.is_file():
                raise RuntimeError(
                    'Failed to verify zephyr_sdk cipd package is readable.'
                )
            result_text = temp_json_path.read_text()
            result_dict = json.loads(result_text)
            if 'result' not in result_dict:
                raise RuntimeError(
                    'Failed to verify zephyr_sdk cipd package is readable.'
                )

        # Initialize cipd
        subprocess.check_call(
            [
                'cipd',
                'init',
                '-force',
                str(core_cache_path),
            ]
        )
        # Install the Zephyr SDK
        subprocess.check_call(
            [
                'cipd',
                'install',
                cipd_package_subpath,
                '-root',
                str(core_cache_path),
                '-force',
            ]
        )
        # Setup Zephyr SDK
        subprocess.check_call(
            [
                'cmake',
                '-P',
                str(core_cache_path / 'cmake' / 'zephyr_sdk_export.cmake'),
            ]
        )

    def install(self, path: pathlib.Path) -> None:
        super().install(path)

        self.__populate_download_cache_from_cipd(path)
        with importlib.resources.path(
            pw_env_setup.virtualenv_setup, 'constraint.list'
        ) as constraint:
            subprocess.check_call(
                [
                    sys.executable,
                    '-m',
                    'pip',
                    'install',
                    '-r',
                    f'{path}/scripts/requirements.txt',
                    '-c',
                    str(constraint),
                ]
            )


pw_package.package_manager.register(Zephyr)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_snapshot_metadata_proto/snapshot_metadata.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_tokenizer_proto import options_pb2 as pw__tokenizer__proto_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2pw_snapshot_metadata_proto/snapshot_metadata.proto\x12\x0bpw.snapshot\x1a pw_tokenizer_proto/options.proto\"\x82\x01\n\x0f\x43puArchitecture\"o\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x41RMV6M\x10\x01\x12\n\n\x06\x41RMV7M\x10\x02\x12\n\n\x06\x41RMV8M\x10\x03\x12\t\n\x05RV32E\x10\x04\x12\t\n\x05RV32I\x10\x05\x12\t\n\x05RV64E\x10\x06\x12\t\n\x05RV64I\x10\x07\x12\n\n\x06RV128I\x10\x08\"\xe9\x01\n\x08Metadata\x12\x14\n\x06reason\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\r\n\x05\x66\x61tal\x18\x02 \x01(\x08\x12\x1a\n\x0cproject_name\x18\x03 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x18\n\x10software_version\x18\x04 \x01(\t\x12\x1b\n\x13software_build_uuid\x18\x05 \x01(\x0c\x12\x19\n\x0b\x64\x65vice_name\x18\x06 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x15\n\rsnapshot_uuid\x18\x07 \x01(\x0c\x12\x33\n\x08\x63pu_arch\x18\x08 \x01(\x0e\x32!.pw.snapshot.CpuArchitecture.Enum\"\xa1\x01\n\x11SnapshotBasicInfo\x12\'\n\x08metadata\x18\x10 \x01(\x0b\x32\x15.pw.snapshot.Metadata\x12\x36\n\x04tags\x18\x11 \x03(\x0b\x32(.pw.snapshot.SnapshotBasicInfo.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x1d\n\x11pw.snapshot.protoB\x08Snapshotb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_snapshot_metadata_proto.snapshot_metadata_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\021pw.snapshot.protoB\010Snapshot'
  _globals['_METADATA'].fields_by_name['reason']._options = None
  _globals['_METADATA'].fields_by_name['reason']._serialized_options = b'\200\257&\000'
  _globals['_METADATA'].fields_by_name['project_name']._options = None
  _globals['_METADATA'].fields_by_name['project_name']._serialized_options = b'\200\257&\000'
  _globals['_METADATA'].fields_by_name['device_name']._options = None
  _globals['_METADATA'].fields_by_name['device_name']._serialized_options = b'\200\257&\000'
  _globals['_SNAPSHOTBASICINFO_TAGSENTRY']._options = None
  _globals['_SNAPSHOTBASICINFO_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_CPUARCHITECTURE']._serialized_start=102
  _globals['_CPUARCHITECTURE']._serialized_end=232
  _globals['_CPUARCHITECTURE_ENUM']._serialized_start=121
  _globals['_CPUARCHITECTURE_ENUM']._serialized_end=232
  _globals['_METADATA']._serialized_start=235
  _globals['_METADATA']._serialized_end=468
  _globals['_SNAPSHOTBASICINFO']._serialized_start=471
  _globals['_SNAPSHOTBASICINFO']._serialized_end=632
  _globals['_SNAPSHOTBASICINFO_TAGSENTRY']._serialized_start=589
  _globals['_SNAPSHOTBASICINFO_TAGSENTRY']._serialized_end=632
# @@protoc_insertion_point(module_scope)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
Helpful commands for working with the repo tool.
See https://gerrit.googlesource.com/git-repo/ for more info!
"""

import subprocess

from typing import List
from pathlib import Path
from pw_presubmit.tools import log_run


def list_all_git_repos() -> List[Path]:
    """Query repo tool and return a list of git repos in the current project.

    Returns:
        List of "Path"s which were found.
    """
    repos = (
        log_run(
            ['repo', 'forall', '-c', 'git', 'rev-parse', '--show-toplevel'],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            check=True,
            ignore_dry_run=True,
        )
        .stdout.decode()
        .strip()
    )

    return [Path(line) for line in repos.splitlines()]



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generate a coverage report using llvm-cov."""

import argparse
import json
import logging
import sys
import subprocess
from pathlib import Path
from typing import Any

_LOG = logging.getLogger(__name__)


def _parser_args() -> dict[str, Any]:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--llvm-cov-path',
        type=Path,
        required=True,
        help='Path to the llvm-cov binary to use to generate coverage reports.',
    )
    parser.add_argument(
        '--format',
        dest='format_type',
        type=str,
        choices=['text', 'html', 'lcov', 'json'],
        required=True,
        help='Desired output format of the code coverage report.',
    )
    parser.add_argument(
        '--test-metadata-path',
        type=Path,
        required=True,
        help='Path to the *.test_metadata.json file that describes all of the '
        'tests being used to generate a coverage report.',
    )
    parser.add_argument(
        '--profdata-path',
        type=Path,
        required=True,
        help='Path for the output merged profdata file to use with generating a'
        ' coverage report for the tests described in --test-metadata.',
    )
    parser.add_argument(
        '--root-dir',
        type=Path,
        required=True,
        help='Path to the project\'s root directory.',
    )
    parser.add_argument(
        '--build-dir',
        type=Path,
        required=True,
        help='Path to the ninja build directory.',
    )
    parser.add_argument(
        '--output-dir',
        type=Path,
        required=True,
        help='Path to where the output report should be placed. This must be a '
        'relative path (from the current working directory) to ensure the '
        'depfiles are generated correctly.',
    )
    parser.add_argument(
        '--depfile-path',
        type=Path,
        required=True,
        help='Path for the output depfile to convey the extra input '
        'requirements from parsing --test-metadata.',
    )
    parser.add_argument(
        '--filter-path',
        dest='filter_paths',
        type=str,
        action='append',
        default=[],
        help='Only these folder paths or files will be included in the output. '
        'To work properly, these must be aboslute paths or relative paths from '
        'the current working directory. No globs or regular expression features'
        ' are supported.',
    )
    parser.add_argument(
        '--ignore-filename-pattern',
        dest='ignore_filename_patterns',
        type=str,
        action='append',
        default=[],
        help='Any file path that matches one of these regular expression '
        'patterns will be excluded from the output report (possibly even if '
        'that path was included in --filter-paths). The regular expression '
        'engine for these is somewhat primitive and does not support things '
        'like look-ahead or look-behind.',
    )
    return vars(parser.parse_args())


def generate_report(
    llvm_cov_path: Path,
    format_type: str,
    test_metadata_path: Path,
    profdata_path: Path,
    root_dir: Path,
    build_dir: Path,
    output_dir: Path,
    depfile_path: Path,
    filter_paths: list[str],
    ignore_filename_patterns: list[str],
) -> int:
    """Generate a coverage report using llvm-cov."""

    # Ensure directories that need to be absolute are.
    root_dir = root_dir.resolve()
    build_dir = build_dir.resolve()

    # Open the test_metadata_path, parse it to JSON, and extract out the
    # test binaries.
    test_metadata = json.loads(test_metadata_path.read_text())
    test_binaries = [
        Path(obj['test_directory']) / obj['test_name']
        for obj in test_metadata
        if 'test_type' in obj and obj['test_type'] == 'unit_test'
    ]

    # llvm-cov export does not create an output file, so we mimic it by creating
    # the directory structure and writing to file outself after we run the
    # command.
    if format_type in ['lcov', 'json']:
        export_output_path = (
            output_dir / 'report.lcov'
            if format_type == 'lcov'
            else output_dir / 'report.json'
        )
        output_dir.mkdir(parents=True, exist_ok=True)

    # Build the command to the llvm-cov subtool based on provided arguments.
    command = [str(llvm_cov_path)]
    if format_type in ['html', 'text']:
        command += [
            'show',
            '--format',
            format_type,
            '--output-dir',
            str(output_dir),
        ]
    else:  # format_type in ['lcov', 'json']
        command += [
            'export',
            '--format',
            # `text` is JSON format when using `llvm-cov`.
            format_type if format_type == 'lcov' else 'text',
        ]
    # We really need two `--path-equivalence` options to be able to map both the
    # root directory for coverage files to the absolute path of the project
    # root_dir and to be able to map "out/" prefix to the provided build_dir.
    #
    # llvm-cov does not currently support two `--path-equivalence` options, so
    # we use `--compilation-dir` and `--path-equivalence` together. This has the
    # unfortunate consequence of showing file paths as absolute in the JSON,
    # LCOV, and text reports.
    #
    # An unwritten assumption here is that root_dir must be an
    # absolute path to enable file-path-based filtering.
    #
    # This is due to turning all file paths into absolute files here:
    # https://github.com/llvm-mirror/llvm/blob/2c4ca6832fa6b306ee6a7010bfb80a3f2596f824/tools/llvm-cov/CodeCoverage.cpp#L188.
    command += [
        '--compilation-dir',
        str(root_dir),
    ]
    # Pigweed maps any build directory to out, which causes generated files to
    # be reported to exist under the out directory, which may not exist if the
    # build directory is not exactly out. This maps out back to the build
    # directory so generated files can be found.
    command += [
        '--path-equivalence',
        f'{str(root_dir)}/out,{str(build_dir)}',
    ]
    command += [
        '--instr-profile',
        str(profdata_path),
    ]
    command += [
        f'--ignore-filename-regex={path}' for path in ignore_filename_patterns
    ]
    # The test binary positional argument MUST appear before the filter path
    # positional arguments. llvm-cov is a horrible interface.
    command += [str(test_binaries[0])]
    command += [f'--object={binary}' for binary in test_binaries[1:]]
    command += [
        str(Path(filter_path).resolve()) for filter_path in filter_paths
    ]

    _LOG.info('')
    _LOG.info(' '.join(command))
    _LOG.info('')

    # Generate the coverage report by invoking the command.
    if format_type in ['html', 'text']:
        output = subprocess.run(command)
        if output.returncode != 0:
            return output.returncode
    else:  # format_type in ['lcov', 'json']
        output = subprocess.run(command, capture_output=True)
        if output.returncode != 0:
            _LOG.error(output.stderr)
            return output.returncode
        export_output_path.write_bytes(output.stdout)

    # Generate the depfile that describes the dependency on the test binaries
    # used to create the report output.
    depfile_target = Path('.')
    if format_type in ['lcov', 'json']:
        depfile_target = export_output_path
    elif format_type == 'text':
        depfile_target = output_dir / 'index.txt'
    else:  # format_type == 'html'
        depfile_target = output_dir / 'index.html'
    depfile_path.write_text(
        ''.join(
            [
                str(depfile_target),
                ': \\\n',
                *[str(binary) + ' \\\n' for binary in test_binaries],
            ]
        )
    )

    return 0


def main() -> int:
    return generate_report(**_parser_args())


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
r"""Decodes and detokenizes strings from binary or Base64 input.

The main class provided by this module is the Detokenize class. To use it,
construct it with the path to an ELF or CSV database, a tokens.Database,
or a file object for an ELF file or CSV. Then, call the detokenize method with
encoded messages, one at a time. The detokenize method returns a
DetokenizedString object with the result.

For example::

  from pw_tokenizer import detokenize

  detok = detokenize.Detokenizer('path/to/firmware/image.elf')
  print(detok.detokenize(b'\x12\x34\x56\x78\x03hi!'))

This module also provides a command line interface for decoding and detokenizing
messages from a file or stdin.
"""

import argparse
import base64
import binascii
from concurrent.futures import Executor, ThreadPoolExecutor
import enum
import io
import logging
import os
from pathlib import Path
import re
import string
import struct
import sys
import threading
import time
from typing import (
    AnyStr,
    BinaryIO,
    Callable,
    Iterable,
    Iterator,
    Match,
    NamedTuple,
    Pattern,
)

try:
    from pw_tokenizer import database, decode, encode, tokens
except ImportError:
    # Append this path to the module search path to allow running this module
    # without installing the pw_tokenizer package.
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    from pw_tokenizer import database, decode, encode, tokens

_LOG = logging.getLogger('pw_tokenizer')

ENCODED_TOKEN = struct.Struct('<I')
_BASE64_CHARS = string.ascii_letters + string.digits + '+/-_='
DEFAULT_RECURSION = 9
NESTED_TOKEN_BASE_PREFIX = encode.NESTED_TOKEN_BASE_PREFIX.encode()

_BASE8_TOKEN_REGEX = rb'(?P<base8>[0-7]{11})'
_BASE10_TOKEN_REGEX = rb'(?P<base10>[0-9]{10})'
_BASE16_TOKEN_REGEX = rb'(?P<base16>[A-Fa-f0-9]{8})'
_BASE64_TOKEN_REGEX = (
    rb'(?P<base64>'
    # Tokenized Base64 contains 0 or more blocks of four Base64 chars.
    rb'(?:[A-Za-z0-9+/\-_]{4})*'
    # The last block of 4 chars may have one or two padding chars (=).
    rb'(?:[A-Za-z0-9+/\-_]{3}=|[A-Za-z0-9+/\-_]{2}==)?'
    rb')'
)
_NESTED_TOKEN_FORMATS = (
    _BASE8_TOKEN_REGEX,
    _BASE10_TOKEN_REGEX,
    _BASE16_TOKEN_REGEX,
    _BASE64_TOKEN_REGEX,
)


def _token_regex(prefix: str) -> Pattern[bytes]:
    """Returns a regular expression for prefixed tokenized strings."""
    return re.compile(
        # Tokenized strings start with the prefix character ($).
        re.escape(prefix.encode())
        # Optional; no base specifier defaults to BASE64.
        # Hash (#) with no number specified defaults to Base-16.
        + rb'(?P<basespec>(?P<base>[0-9]*)?'
        + NESTED_TOKEN_BASE_PREFIX
        + rb')?'
        # Match one of the following token formats.
        + rb'('
        + rb'|'.join(_NESTED_TOKEN_FORMATS)
        + rb')'
    )


class DetokenizedString:
    """A detokenized string, with all results if there are collisions."""

    def __init__(
        self,
        token: int | None,
        format_string_entries: Iterable[tuple],
        encoded_message: bytes,
        show_errors: bool = False,
        recursive_detokenize: Callable[[str], str] | None = None,
    ):
        self.token = token
        self.encoded_message = encoded_message
        self._show_errors = show_errors

        self.successes: list[decode.FormattedString] = []
        self.failures: list[decode.FormattedString] = []

        decode_attempts: list[tuple[tuple, decode.FormattedString]] = []

        for entry, fmt in format_string_entries:
            result = fmt.format(
                encoded_message[ENCODED_TOKEN.size :], show_errors
            )
            if recursive_detokenize:
                result = decode.FormattedString(
                    recursive_detokenize(result.value),
                    result.args,
                    result.remaining,
                )
            decode_attempts.append((result.score(entry.date_removed), result))

        # Sort the attempts by the score so the most likely results are first.
        decode_attempts.sort(key=lambda value: value[0], reverse=True)

        # Split out the successesful decodes from the failures.
        for score, result in decode_attempts:
            if score[0]:
                self.successes.append(result)
            else:
                self.failures.append(result)

    def ok(self) -> bool:
        """True if exactly one string decoded the arguments successfully."""
        return len(self.successes) == 1

    def matches(self) -> list[decode.FormattedString]:
        """Returns the strings that matched the token, best matches first."""
        return self.successes + self.failures

    def best_result(self) -> decode.FormattedString | None:
        """Returns the string and args for the most likely decoded string."""
        for string_and_args in self.matches():
            return string_and_args

        return None

    def error_message(self) -> str:
        """If detokenization failed, returns a descriptive message."""
        if self.ok():
            return ''

        if not self.matches():
            if self.token is None:
                return 'missing token'

            return 'unknown token {:08x}'.format(self.token)

        if len(self.matches()) == 1:
            return 'decoding failed for {!r}'.format(self.matches()[0].value)

        return '{} matches'.format(len(self.matches()))

    def __str__(self) -> str:
        """Returns the string for the most likely result."""
        result = self.best_result()
        if result:
            return result[0]

        if self._show_errors:
            return '<[ERROR: {}|{!r}]>'.format(
                self.error_message(), self.encoded_message
            )

        # Display the string as prefixed Base64 if it cannot be decoded.
        return encode.prefixed_base64(self.encoded_message)

    def __repr__(self) -> str:
        if self.ok():
            message = repr(str(self))
        else:
            message = 'ERROR: {}|{!r}'.format(
                self.error_message(), self.encoded_message
            )

        return '{}({})'.format(type(self).__name__, message)


class _TokenizedFormatString(NamedTuple):
    entry: tokens.TokenizedStringEntry
    format: decode.FormatString


class Detokenizer:
    """Main detokenization class; detokenizes strings and caches results."""

    def __init__(
        self,
        *token_database_or_elf,
        show_errors: bool = False,
        prefix: str | bytes = encode.NESTED_TOKEN_PREFIX,
    ):
        """Decodes and detokenizes binary messages.

        Args:
          *token_database_or_elf: a path or file object for an ELF or CSV
              database, a tokens.Database, or an elf_reader.Elf
          prefix: one-character byte string that signals the start of a message
          show_errors: if True, an error message is used in place of the %
              conversion specifier when an argument fails to decode
        """
        self.show_errors = show_errors
        self._prefix = prefix if isinstance(prefix, str) else prefix.decode()
        self._token_regex = _token_regex(self._prefix)

        self._database_lock = threading.Lock()

        # Cache FormatStrings for faster lookup & formatting.
        self._cache: dict[int, list[_TokenizedFormatString]] = {}

        self._initialize_database(token_database_or_elf)

    @property
    def prefix(self) -> str:
        return self._prefix

    def _initialize_database(self, token_sources: Iterable) -> None:
        with self._database_lock:
            self.database = database.load_token_database(*token_sources)
            self._cache.clear()

    def lookup(self, token: int) -> list[_TokenizedFormatString]:
        """Returns (TokenizedStringEntry, FormatString) list for matches."""
        with self._database_lock:
            try:
                return self._cache[token]
            except KeyError:
                format_strings = [
                    _TokenizedFormatString(
                        entry, decode.FormatString(str(entry))
                    )
                    for entry in self.database.token_to_entries[token]
                ]
                self._cache[token] = format_strings
                return format_strings

    def detokenize(
        self,
        encoded_message: bytes,
        recursion: int = DEFAULT_RECURSION,
    ) -> DetokenizedString:
        """Decodes and detokenizes a message as a DetokenizedString."""
        if not encoded_message:
            return DetokenizedString(
                None, (), encoded_message, self.show_errors
            )

        # Pad messages smaller than ENCODED_TOKEN.size with zeroes to support
        # tokens smaller than a uint32. Messages with arguments must always use
        # a full 32-bit token.
        missing_token_bytes = ENCODED_TOKEN.size - len(encoded_message)
        if missing_token_bytes > 0:
            encoded_message += b'\0' * missing_token_bytes

        (token,) = ENCODED_TOKEN.unpack_from(encoded_message)

        recursive_detokenize = None
        if recursion > 0:
            recursive_detokenize = self._detokenize_nested_callback(recursion)

        return DetokenizedString(
            token,
            self.lookup(token),
            encoded_message,
            self.show_errors,
            recursive_detokenize,
        )

    def detokenize_text(
        self,
        data: AnyStr,
        recursion: int = DEFAULT_RECURSION,
    ) -> AnyStr:
        """Decodes and replaces prefixed Base64 messages in the provided data.

        Args:
          data: the binary data to decode
          recursion: how many levels to recursively decode

        Returns:
          copy of the data with all recognized tokens decoded
        """
        return self._detokenize_nested_callback(recursion)(data)

    # TODO(gschen): remove unnecessary function
    def detokenize_base64(
        self,
        data: AnyStr,
        recursion: int = DEFAULT_RECURSION,
    ) -> AnyStr:
        """Alias of detokenize_text for backwards compatibility."""
        return self.detokenize_text(data, recursion)

    def detokenize_text_to_file(
        self,
        data: AnyStr,
        output: BinaryIO,
        recursion: int = DEFAULT_RECURSION,
    ) -> None:
        """Decodes prefixed Base64 messages in data; decodes to output file."""
        output.write(self._detokenize_nested(data, recursion))

    # TODO(gschen): remove unnecessary function
    def detokenize_base64_to_file(
        self,
        data: AnyStr,
        output: BinaryIO,
        recursion: int = DEFAULT_RECURSION,
    ) -> None:
        """Alias of detokenize_text_to_file for backwards compatibility."""
        self.detokenize_text_to_file(data, output, recursion)

    def detokenize_text_live(
        self,
        input_file: io.RawIOBase | BinaryIO,
        output: BinaryIO,
        recursion: int = DEFAULT_RECURSION,
    ) -> None:
        """Reads chars one-at-a-time, decoding messages; SLOW for big files."""

        def transform(data: bytes) -> bytes:
            return self._detokenize_nested(data.decode(), recursion)

        for message in NestedMessageParser(
            self._prefix, _BASE64_CHARS
        ).transform_io(input_file, transform):
            output.write(message)

            # Flush each line to prevent delays when piping between processes.
            if b'\n' in message:
                output.flush()

    # TODO(gschen): remove unnecessary function
    def detokenize_base64_live(
        self,
        input_file: io.RawIOBase | BinaryIO,
        output: BinaryIO,
        recursion: int = DEFAULT_RECURSION,
    ) -> None:
        """Alias of detokenize_text_live for backwards compatibility."""
        self.detokenize_text_live(input_file, output, recursion)

    def _detokenize_nested_callback(
        self,
        recursion: int,
    ) -> Callable[[AnyStr], AnyStr]:
        """Returns a function that replaces all tokens for a given string."""

        def detokenize(message: AnyStr) -> AnyStr:
            result = self._detokenize_nested(message, recursion)
            return result.decode() if isinstance(message, str) else result

        return detokenize

    def _detokenize_nested(
        self,
        message: str | bytes,
        recursion: int,
    ) -> bytes:
        """Returns the message with recognized tokens replaced.

        Message data is internally handled as bytes regardless of input message
        type and returns the result as bytes.
        """
        # A unified format across the token types is required for regex
        # consistency.
        message = message.encode() if isinstance(message, str) else message

        if not self.database:
            return message

        result = message
        for _ in range(recursion - 1):
            result = self._token_regex.sub(self._detokenize_scan, result)

            if result == message:
                return result
        return result

    def _detokenize_scan(self, match: Match[bytes]) -> bytes:
        """Decodes prefixed tokens for one of multiple formats."""
        basespec = match.group('basespec')
        base = match.group('base')

        if not basespec or (base == b'64'):
            return self._detokenize_once_base64(match)

        if not base:
            base = b'16'

        return self._detokenize_once(match, base)

    def _detokenize_once(
        self,
        match: Match[bytes],
        base: bytes,
    ) -> bytes:
        """Performs lookup on a plain token"""
        original = match.group(0)
        token = match.group('base' + base.decode())
        if not token:
            return original

        token = int(token, int(base))
        entries = self.database.token_to_entries[token]

        if len(entries) == 1:
            return str(entries[0]).encode()

        # TODO(gschen): improve token collision reporting

        return original

    def _detokenize_once_base64(
        self,
        match: Match[bytes],
    ) -> bytes:
        """Performs lookup on a Base64 token"""
        original = match.group(0)

        try:
            encoded_token = match.group('base64')
            if not encoded_token:
                return original

            detokenized_string = self.detokenize(
                base64.b64decode(encoded_token, validate=True), recursion=0
            )

            if detokenized_string.matches():
                return str(detokenized_string).encode()

        except binascii.Error:
            pass

        return original


# TODO: b/265334753 - Reuse this function in database.py:LoadTokenDatabases
def _parse_domain(path: Path | str) -> tuple[Path, Pattern[str] | None]:
    """Extracts an optional domain regex pattern suffix from a path"""

    if isinstance(path, Path):
        path = str(path)

    delimiters = path.count('#')

    if delimiters == 0:
        return Path(path), None

    if delimiters == 1:
        path, domain = path.split('#')
        return Path(path), re.compile(domain)

    raise ValueError(
        f'Too many # delimiters. Expected 0 or 1, found {delimiters}'
    )


class AutoUpdatingDetokenizer(Detokenizer):
    """Loads and updates a detokenizer from database paths."""

    class _DatabasePath:
        """Tracks the modified time of a path or file object."""

        def __init__(self, path: Path | str) -> None:
            self.path, self.domain = _parse_domain(path)
            self._modified_time: float | None = self._last_modified_time()

        def updated(self) -> bool:
            """True if the path has been updated since the last call."""
            modified_time = self._last_modified_time()
            if modified_time is None or modified_time == self._modified_time:
                return False

            self._modified_time = modified_time
            return True

        def _last_modified_time(self) -> float | None:
            if self.path.is_dir():
                mtime = -1.0
                for child in self.path.glob(tokens.DIR_DB_GLOB):
                    mtime = max(mtime, os.path.getmtime(child))
                return mtime if mtime >= 0 else None

            try:
                return os.path.getmtime(self.path)
            except FileNotFoundError:
                return None

        def load(self) -> tokens.Database:
            try:
                if self.domain is not None:
                    return database.load_token_database(
                        self.path, domain=self.domain
                    )
                return database.load_token_database(self.path)
            except FileNotFoundError:
                return database.load_token_database()

    def __init__(
        self,
        *paths_or_files: Path | str,
        min_poll_period_s: float = 1.0,
        pool: Executor = ThreadPoolExecutor(max_workers=1),
    ) -> None:
        self.paths = tuple(self._DatabasePath(path) for path in paths_or_files)
        self.min_poll_period_s = min_poll_period_s
        self._last_checked_time: float = time.time()
        # Thread pool to use for loading the databases. Limit to a single
        # worker since this is low volume and not time critical.
        self._pool = pool
        super().__init__(*(path.load() for path in self.paths))

    def __del__(self) -> None:
        self._pool.shutdown(wait=False)

    def _reload_paths(self) -> None:
        self._initialize_database([path.load() for path in self.paths])

    def _reload_if_changed(self) -> None:
        if time.time() - self._last_checked_time >= self.min_poll_period_s:
            self._last_checked_time = time.time()

            if any(path.updated() for path in self.paths):
                _LOG.info('Changes detected; reloading token database')
                self._pool.submit(self._reload_paths)

    def lookup(self, token: int) -> list[_TokenizedFormatString]:
        self._reload_if_changed()
        return super().lookup(token)


class NestedMessageParser:
    """Parses nested tokenized messages from a byte stream or string."""

    class _State(enum.Enum):
        MESSAGE = 1
        NON_MESSAGE = 2

    def __init__(
        self,
        prefix: str | bytes = encode.NESTED_TOKEN_PREFIX,
        chars: str | bytes = _BASE64_CHARS,
    ) -> None:
        """Initializes a parser.

        Args:
            prefix: one character that signifies the start of a message (``$``).
            chars: characters allowed in a message
        """
        self._prefix = ord(prefix)

        if isinstance(chars, str):
            chars = chars.encode()

        # Store the valid message bytes as a set of byte values.
        self._message_bytes = frozenset(chars)

        if len(prefix) != 1 or self._prefix in self._message_bytes:
            raise ValueError(
                f'Invalid prefix {prefix!r}: the prefix must be a single '
                'character that is not a valid message character.'
            )

        self._buffer = bytearray()
        self._state: NestedMessageParser._State = self._State.NON_MESSAGE

    def read_messages_io(
        self, binary_io: io.RawIOBase | BinaryIO
    ) -> Iterator[tuple[bool, bytes]]:
        """Reads prefixed messages from a byte stream (BinaryIO object).

        Reads until EOF. If the stream is nonblocking (``read(1)`` returns
        ``None``), then this function returns and may be called again with the
        same IO object to continue parsing. Partial messages are preserved
        between calls.

        Yields:
            ``(is_message, contents)`` chunks.
        """
        # The read may block indefinitely, depending on the IO object.
        while (read_byte := binary_io.read(1)) != b'':
            # Handle non-blocking IO by returning when no bytes are available.
            if read_byte is None:
                return

            for byte in read_byte:
                yield from self._handle_byte(byte)

            if self._state is self._State.NON_MESSAGE:  # yield non-message byte
                yield from self._flush()

        yield from self._flush()  # Always flush after EOF
        self._state = self._State.NON_MESSAGE

    def read_messages(
        self, chunk: bytes, *, flush: bool = False
    ) -> Iterator[tuple[bool, bytes]]:
        """Reads prefixed messages from a byte string.

        This function may be called repeatedly with chunks of a stream. Partial
        messages are preserved between calls, unless ``flush=True``.

        Args:
            chunk: byte string that may contain nested messagses
            flush: whether to flush any incomplete messages after processing
                this chunk

        Yields:
            ``(is_message, contents)`` chunks.
        """
        for byte in chunk:
            yield from self._handle_byte(byte)

        if flush or self._state is self._State.NON_MESSAGE:
            yield from self._flush()

    def _handle_byte(self, byte: int) -> Iterator[tuple[bool, bytes]]:
        if self._state is self._State.MESSAGE:
            if byte not in self._message_bytes:
                yield from self._flush()
                if byte != self._prefix:
                    self._state = self._State.NON_MESSAGE
        elif self._state is self._State.NON_MESSAGE:
            if byte == self._prefix:
                yield from self._flush()
                self._state = self._State.MESSAGE
        else:
            raise NotImplementedError(f'Unsupported state: {self._state}')

        self._buffer.append(byte)

    def _flush(self) -> Iterator[tuple[bool, bytes]]:
        data = bytes(self._buffer)
        self._buffer.clear()
        if data:
            yield self._state is self._State.MESSAGE, data

    def transform_io(
        self,
        binary_io: io.RawIOBase | BinaryIO,
        transform: Callable[[bytes], bytes],
    ) -> Iterator[bytes]:
        """Yields the file with a transformation applied to the messages."""
        for is_message, chunk in self.read_messages_io(binary_io):
            yield transform(chunk) if is_message else chunk

    def transform(
        self,
        chunk: bytes,
        transform: Callable[[bytes], bytes],
        *,
        flush: bool = False,
    ) -> bytes:
        """Yields the chunk with a transformation applied to the messages.

        Partial messages are preserved between calls unless ``flush=True``.
        """
        return b''.join(
            transform(data) if is_message else data
            for is_message, data in self.read_messages(chunk, flush=flush)
        )


# TODO(hepler): Remove this unnecessary function.
def detokenize_base64(
    detokenizer: Detokenizer,
    data: bytes,
    recursion: int = DEFAULT_RECURSION,
) -> bytes:
    """Alias for detokenizer.detokenize_base64 for backwards compatibility.

    This function is deprecated; do not call it.
    """
    return detokenizer.detokenize_base64(data, recursion)


def _follow_and_detokenize_file(
    detokenizer: Detokenizer,
    file: BinaryIO,
    output: BinaryIO,
    poll_period_s: float = 0.01,
) -> None:
    """Polls a file to detokenize it and any appended data."""

    try:
        while True:
            data = file.read()
            if data:
                detokenizer.detokenize_base64_to_file(data, output)
                output.flush()
            else:
                time.sleep(poll_period_s)
    except KeyboardInterrupt:
        pass


def _handle_base64(
    databases,
    input_file: BinaryIO,
    output: BinaryIO,
    prefix: str,
    show_errors: bool,
    follow: bool,
) -> None:
    """Handles the base64 command line option."""
    # argparse.FileType doesn't correctly handle - for binary files.
    if input_file is sys.stdin:
        input_file = sys.stdin.buffer

    if output is sys.stdout:
        output = sys.stdout.buffer

    detokenizer = Detokenizer(
        tokens.Database.merged(*databases),
        prefix=prefix,
        show_errors=show_errors,
    )

    if follow:
        _follow_and_detokenize_file(detokenizer, input_file, output)
    elif input_file.seekable():
        # Process seekable files all at once, which is MUCH faster.
        detokenizer.detokenize_base64_to_file(input_file.read(), output)
    else:
        # For non-seekable inputs (e.g. pipes), read one character at a time.
        detokenizer.detokenize_base64_live(input_file, output)


def _parse_args() -> argparse.Namespace:
    """Parses and return command line arguments."""

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.set_defaults(handler=lambda **_: parser.print_help())

    subparsers = parser.add_subparsers(help='Encoding of the input.')

    base64_help = 'Detokenize Base64-encoded data from a file or stdin.'
    subparser = subparsers.add_parser(
        'base64',
        description=base64_help,
        parents=[database.token_databases_parser()],
        help=base64_help,
    )
    subparser.set_defaults(handler=_handle_base64)
    subparser.add_argument(
        '-i',
        '--input',
        dest='input_file',
        type=argparse.FileType('rb'),
        default=sys.stdin.buffer,
        help='The file from which to read; provide - or omit for stdin.',
    )
    subparser.add_argument(
        '-f',
        '--follow',
        action='store_true',
        help=(
            'Detokenize data appended to input_file as it grows; similar to '
            'tail -f.'
        ),
    )
    subparser.add_argument(
        '-o',
        '--output',
        type=argparse.FileType('wb'),
        default=sys.stdout.buffer,
        help=(
            'The file to which to write the output; '
            'provide - or omit for stdout.'
        ),
    )
    subparser.add_argument(
        '-p',
        '--prefix',
        default=encode.NESTED_TOKEN_PREFIX,
        help=(
            'The one-character prefix that signals the start of a '
            'nested tokenized message. (default: $)'
        ),
    )
    subparser.add_argument(
        '-s',
        '--show_errors',
        action='store_true',
        help=(
            'Show error messages instead of conversion specifiers when '
            'arguments cannot be decoded.'
        ),
    )

    return parser.parse_args()


def main() -> int:
    args = _parse_args()

    handler = args.handler
    del args.handler

    handler(**vars(args))
    return 0


if __name__ == '__main__':
    if sys.version_info[0] < 3:
        sys.exit('ERROR: The detokenizer command line tools require Python 3.')
    sys.exit(main())



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Sensor schema validation tooling."""

from collections.abc import Sequence
import importlib.resources
import logging
from pathlib import Path

import jsonschema  # type: ignore
import jsonschema.exceptions  # type: ignore
import yaml

_METADATA_SCHEMA = yaml.safe_load(
    importlib.resources.read_text("pw_sensor", "metadata_schema.json")
)

_DEPENDENCY_SCHEMA = yaml.safe_load(
    importlib.resources.read_text("pw_sensor", "dependency_schema.json")
)

_RESOLVED_SCHEMA = yaml.safe_load(
    importlib.resources.read_text("pw_sensor", "resolved_schema.json")
)


class Validator:
    """
    Context used for validating metadata dictionaries.

    What the validator is:
    - A system to resolve and verify that declared sensor metadata is well
      defined and formatted
    - A utility to resolve any and all dependencies when using a specified
      metadata file.

    What the validator is NOT:
    - Code generator
    """

    def __init__(
        self,
        include_paths: Sequence[Path] | None = None,
        log_level: int = logging.WARNING,
    ) -> None:
        """
        Construct a Validator with some context of the current run.

        Args:
          include_paths: An optional list of directories in which to resolve
            dependencies
          log_level: A desired logging level (defaults to logging.WARNING)
        """
        self._include_paths = include_paths if include_paths else []
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(log_level)

    def validate(self, metadata: dict) -> dict:
        """
        Accept a structured metadata description. This dictionary should first
        pass the schema provided in metadata_schema.yaml. Then, every channel
        used by the sensor should be defined in exactly one of the dependencies.
        Example YAML:

          deps:
            - "pw_sensor/channels.yaml"
          compatible:
            org: "Bosch"
            part: "BMA4xx
          supported-buses:
            - i2c
          channels:
            acceleration: []
            die_temperature: []

        Args:
          metadata: Structured sensor data, this will NOT be modified

        Returns:
          A set of attributes, channels, triggers, and units along with a single
          sensor which match the schema in resolved_schema.json.

        Raises:
          RuntimeError: An error in the schema validation or a missing
            definition.
          FileNotFoundError: One of the dependencies was not found.
        """
        result: dict = {
            "attributes": {},
            "channels": {},
            "triggers": {},
            "units": {},
            "sensors": {},
        }
        metadata = metadata.copy()

        # Validate the incoming schema
        try:
            jsonschema.validate(instance=metadata, schema=_METADATA_SCHEMA)
        except jsonschema.exceptions.ValidationError as e:
            raise RuntimeError(
                "ERROR: Malformed sensor metadata YAML:\n"
                f"{yaml.safe_dump(metadata, indent=2)}"
            ) from e

        # Resolve all the dependencies, after this, 'result' will have all the
        # missing properties for which defaults can be provided
        self._resolve_dependencies(metadata=metadata, out=result)

        self._logger.debug(
            "Resolved dependencies:\n%s", yaml.safe_dump(result, indent=2)
        )

        # Resolve all channel entries
        self._resolve_channels(metadata=metadata, out=result)

        # Resolve all attribute entries
        self._resolve_attributes(metadata=metadata, out=result)

        # Resolve all trigger entries
        self._resolve_triggers(metadata=metadata, out=result)

        compatible = metadata.pop("compatible")
        supported_buses = metadata.pop("supported-buses")
        channels = metadata.pop("channels")
        attributes = metadata.pop("attributes")
        triggers = metadata.pop("triggers")
        result["sensors"][f"{compatible['org']},{compatible['part']}"] = {
            "compatible": compatible,
            "supported-buses": supported_buses,
            "channels": channels,
            "attributes": attributes,
            "triggers": triggers,
            "description": metadata.get("description", ""),
        }

        # Validate the final output before returning
        try:
            jsonschema.validate(instance=result, schema=_RESOLVED_SCHEMA)
        except jsonschema.exceptions.ValidationError as e:
            raise RuntimeError(
                "ERROR: Malformed output YAML: "
                f"{yaml.safe_dump(result, indent=2)}"
            ) from e

        return result

    def _resolve_dependencies(self, metadata: dict, out: dict) -> None:
        """
        Given a list of dependencies, ensure that each of them exists and
        matches the schema provided in dependency_schema.yaml. Once loaded, the
        content of the definition file will be resolved (filling in any missing
        fields that can be inherited) and the final result will be placed in the
        'out' dictionary.

        Args:
          metadata: The full sensor metadata passed to the validate function
          out: Output dictionary where the resolved dependencies should be
            stored

        Raises:
          RuntimeError: An error in the schema validation or a missing
            definition.
          FileNotFoundError: One of the dependencies was not found.
        """
        deps: None | list[str] = metadata.get("deps")
        if not deps:
            self._logger.debug("No dependencies found, skipping imports")
            return

        merged_deps: dict = {
            "attributes": {},
            "channels": {},
            "triggers": {},
            "units": {},
        }
        for dep in deps:
            # Load each of the dependencies, then merge them. This avoids any
            # include dependency order issues.
            dep_file = self._get_dependency_file(dep)
            with open(dep_file, mode="r", encoding="utf-8") as dep_yaml_file:
                dep_yaml = yaml.safe_load(dep_yaml_file)
                try:
                    jsonschema.validate(
                        instance=dep_yaml, schema=_DEPENDENCY_SCHEMA
                    )
                except jsonschema.exceptions.ValidationError as e:
                    raise RuntimeError(
                        "ERROR: Malformed dependency YAML: "
                        f"{yaml.safe_dump(dep_yaml, indent=2)}"
                    ) from e
                # Merge all the loaded values into 'merged_deps'
                for category in merged_deps:
                    self._merge_deps(
                        category=category,
                        dep_yaml=dep_yaml,
                        merged_deps=merged_deps,
                    )
        # Backfill any default values from the merged dependencies and put them
        # into 'out'
        self._backfill_declarations(declarations=merged_deps, out=out)

    @staticmethod
    def _merge_deps(category: str, dep_yaml: dict, merged_deps: dict) -> None:
        """
        Pull all properties from dep_yaml[category] and put them into
        merged_deps after validating that no key duplicates exist.

        Args:
          category: The index of dep_yaml and merged_deps to merge
          dep_yaml: The newly loaded dependency YAML
          merged_deps: The accumulated dependency map
        """
        for key, value in dep_yaml.get(category, {}).items():
            assert (
                key not in merged_deps[category]
            ), f"'{key}' was already found under '{category}'"
            merged_deps[category][key] = value

    def _backfill_declarations(self, declarations: dict, out: dict) -> None:
        """
        Add any missing properties of a declaration object.

        Args:
          declarations: The top level declarations dictionary loaded from the
            dependency file.
          out: The already resolved map of all defined dependencies
        """
        self._backfill_units(declarations=declarations, out=out)
        self._backfill_channels(declarations=declarations, out=out)
        self._backfill_attributes(declarations=declarations, out=out)
        self._backfill_triggers(declarations=declarations, out=out)

    @staticmethod
    def _backfill_units(declarations: dict, out: dict) -> None:
        """
        Move units from 'declarations' to 'out' while also filling in any
        default values.

        Args:
          declarations: The original YAML declaring units.
          out: Output dictionary where we'll add the key "units" wit the result.
        """
        if out.get("units") is None:
            out["units"] = {}
        resolved_units: dict = out["units"]
        if not declarations.get("units"):
            return

        for units_id, unit in declarations["units"].items():
            # Copy unit to resolved_units and fill any default values
            assert resolved_units.get(units_id) is None
            resolved_units[units_id] = unit
            if not unit.get("name"):
                unit["name"] = unit["symbol"]
            if unit.get("description") is None:
                unit["description"] = ""

    @staticmethod
    def _backfill_attributes(declarations: dict, out: dict) -> None:
        """
        Move attributes from 'delcarations' to 'out' while also filling in any
        default values.

        Args:
          declarations: The original YAML declaring attributes.
          out: Output dictionary where we'll add the key "attributes" with the
            result.
        """
        if out.get("attributes") is None:
            out["attributes"] = {}
        resolved_attributes: dict = out["attributes"]
        if not declarations.get("attributes"):
            return

        for attr_id, attribute in declarations["attributes"].items():
            # Copy attribute to resolved_attributes and fill any default values
            assert resolved_attributes.get(attr_id) is None
            resolved_attributes[attr_id] = attribute
            if not attribute.get("name"):
                attribute["name"] = attr_id
            if not attribute.get("description"):
                attribute["description"] = ""

    @staticmethod
    def _backfill_channels(declarations: dict, out: dict) -> None:
        """
        Move channels from 'declarations' to 'out' while also filling in any
        default values.

        Args:
          declarations: The original YAML declaring channels.
          out: Output dictionary where we'll add the key "channels" with the
            result.
        """
        if out.get("channels") is None:
            out["channels"] = {}
        resolved_channels: dict = out["channels"]
        if not declarations.get("channels"):
            return

        for chan_id, channel in declarations["channels"].items():
            # Copy channel to resolved_channels and fill any default values
            assert resolved_channels.get(chan_id) is None
            resolved_channels[chan_id] = channel
            if not channel.get("name"):
                channel["name"] = chan_id
            if not channel.get("description"):
                channel["description"] = ""
            assert channel["units"] in out["units"], (
                f"'{channel['units']}' not found in\n"
                + f"{yaml.safe_dump(out.get('units', {}), indent=2)}"
            )

    @staticmethod
    def _backfill_triggers(declarations: dict, out: dict) -> None:
        """
        Move triggers from 'delcarations' to 'out' while also filling in any
        default values.

        Args:
          declarations: The original YAML declaring triggers.
          out: Output dictionary where we'll add the key "triggers" with the
            result.
        """
        if out.get("triggers") is None:
            out["triggers"] = {}
        resolved_triggers: dict = out["triggers"]
        if not declarations.get("triggers"):
            return

        for trigger_id, trigger in declarations["triggers"].items():
            # Copy trigger to resolved_triggers and fill any default values
            assert resolved_triggers.get(trigger_id) is None
            resolved_triggers[trigger_id] = trigger
            if not trigger.get("name"):
                trigger["name"] = trigger_id
            if not trigger.get("description"):
                trigger["description"] = ""

    def _resolve_attributes(self, metadata: dict, out: dict) -> None:
        """
        For each attribute in the metadta, find the matching definition in the
        'out/attributes' entry and use the data to fill any missing information.
        For example, if an entry exists that looks like:
            sample_rate: {}

        We would then try and find the 'sample_rate' key in the out/attributes
        list (which was already validated by _resolve_dependencies). Since the
        example above does not override any fields, we would copy the 'name',
        'description', and 'units' from the definition into the attribute entry.

        Args:
          metadata: The full sensor metadata passed to the validate function
          out: The current output, used to get channel definitions

        Raises:
          RuntimeError: An error in the schema validation or a missing
            definition.
        """
        attributes: list | None = metadata.get("attributes")
        if not attributes:
            metadata["attributes"] = []
            self._logger.debug("No attributes found, skipping")
            return

        attribute: dict
        for attribute in attributes:
            assert attribute["attribute"] in out["attributes"]
            assert attribute["channel"] in out["channels"]
            assert attribute["units"] in out["units"]

    def _resolve_channels(self, metadata: dict, out: dict) -> None:
        """
        For each channel in the metadata, find the matching definition in the
        'out/channels' entry and use the data to fill any missing information.
        For example, if an entry exists that looks like:
            acceleration: {}

        We would then try and find the 'acceleration' key in the out/channels
        dict (which was already validated by _resolve_dependencies). Since the
        example above does not override any fields, we would copy the 'name',
        'description', and 'units' from the definition into the channel entry.

        Args:
          metadata: The full sensor metadata passed to the validate function
          out: The current output, used to get channel definitions

        Raises:
          RuntimeError: An error in the schema validation or a missing
            definition.
        """
        channels: dict | None = metadata.get("channels")
        if not channels:
            self._logger.debug("No channels found, skipping")
            metadata["channels"] = {}
            return

        channel_name: str
        indices: list[dict]
        for channel_name, indices in channels.items():
            # channel_name must have been resolved by now.
            if out["channels"].get(channel_name) is None:
                raise RuntimeError(
                    f"Failed to find a definition for '{channel_name}', did you"
                    " forget a dependency?"
                )
            channel = out["channels"][channel_name]
            # The content of 'channel' came from the 'out/channels' dict which
            # was already validated and every field added if missing. At this
            # point it's safe to access the channel's name, description, and
            # units.

            if not indices:
                indices.append({})

            index: dict
            for index in indices:
                if not index.get("name"):
                    index["name"] = channel["name"]
                if not index.get("description"):
                    index["description"] = channel["description"]
                # Always use the same units
                index["units"] = channel["units"]

    def _resolve_triggers(self, metadata: dict, out: dict) -> None:
        """
        For each trigger in the metadata, find the matching definition in the
        'out/triggers' entry and use the data to fill any missing information.
        For example, if an entry exists that looks like:
            data_ready: {}

        We would then try and find the 'data_ready' key in the out/triggers
        dict (which was already validated by _resolve_dependencies). Since the
        example above does not override any fields, we would copy the 'name' and
        'description' from the definition into the trigger entry.

        Args:
          metadata: The full sensor metadata passed to the validate function
          out: The current output, used to get trigger definitions

        Raises:
          RuntimeError: An error in the schema validation or a missing
            definition.
        """
        triggers: list | None = metadata.get("triggers")
        if not triggers:
            metadata["triggers"] = []
            self._logger.debug("No triggers found, skipping")
            return

        for trigger_name in triggers:
            assert trigger_name in out["triggers"]

    def _get_dependency_file(self, dep: str) -> Path:
        """
        Search for a dependency file and return the full path to it if found.

        Args:
          dep: The dependency string as provided by the metadata yaml.

        Returns:
          The dependency file as a Path object if found.

        Raises:
          FileNotFoundError: One of the dependencies was not found.
        """
        error_string = f"Failed to find {dep} using search paths:"
        # Check if a full path was used
        if Path(dep).is_file():
            return Path(dep)

        # Search all the include paths
        for path in self._include_paths:
            if (path / dep).is_file():
                return path / dep
            error_string += f"\n- {path}"

        raise FileNotFoundError(error_string)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Reads and parses the Pigweed config file.

See also https://pigweed.dev/seed/0101-pigweed.json.html.
"""

import json
import os

try:
    # This will only succeed when using config_file from Bazel.
    from python.runfiles import runfiles  # type: ignore
except (ImportError, ModuleNotFoundError):
    runfiles = None


def _resolve_env(env):
    if env:
        return env
    return os.environ


def _get_project_root(env):
    for var in ('PW_PROJECT_ROOT', 'PW_ROOT'):
        if var in env:
            return env[var]
    raise ValueError('environment variable PW_PROJECT_ROOT not set')


def _pw_env_substitute(env, string):
    if not isinstance(string, str):
        return string

    # Substitute in environment variables based on $pw_env{VAR_NAME} tokens.
    for key, value in env.items():
        string = string.replace('$pw_env{' + key + '}', value)

    if '$pw_env{' in string:
        raise ValueError(f'Unresolved $pw_env\\{...} in JSON string: {string}')

    return string


def path(env=None):
    """Return the path where pigweed.json should reside."""
    env = _resolve_env(env)
    return os.path.join(_get_project_root(env=env), 'pigweed.json')


def path_from_runfiles():
    r = runfiles.Create()
    location = r.Rlocation("pigweed.json")
    if location is None:
        # Failed to find pigweed.json
        raise ValueError("Failed to find pigweed.json")

    return location


def load(env=None):
    """Load pigweed.json if it exists and return the contents."""
    if env is None and runfiles is not None:
        config_path = path_from_runfiles()
    else:
        env = _resolve_env(env)
        config_path = path(env=env)

    if not os.path.isfile(config_path):
        return {}

    def hook(obj):
        out = {}
        for key, val in obj.items():
            out[key] = _pw_env_substitute(env, val)
        return out

    with open(config_path, 'r') as file:
        return json.load(file, object_hook=hook)



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities for using HDLC with ``pw_rpc``."""

from __future__ import annotations

import io
import logging
import queue
import sys
import threading
import time
from typing import (
    Any,
    BinaryIO,
    Callable,
    Iterable,
    Sequence,
    TypeVar,
)

import pw_rpc
from pw_rpc import client_utils
from pw_hdlc.decode import Frame, FrameDecoder
from pw_hdlc import encode
from pw_stream import stream_readers

_LOG = logging.getLogger('pw_hdlc.rpc')
_VERBOSE = logging.DEBUG - 1


# Aliases for objects moved to their proper place in pw_rpc formerly defined
# here. This is for backwards compatibility.
CancellableReader = stream_readers.CancellableReader
SelectableReader = stream_readers.SelectableReader
SocketReader = stream_readers.SocketReader
SerialReader = stream_readers.SerialReader
DataReaderAndExecutor = stream_readers.DataReaderAndExecutor
PathsModulesOrProtoLibrary = client_utils.PathsModulesOrProtoLibrary
RpcClient = client_utils.RpcClient
NoEncodingSingleChannelRpcClient = client_utils.NoEncodingSingleChannelRpcClient
SocketSubprocess = stream_readers.SocketSubprocess
FrameTypeT = TypeVar('FrameTypeT')


# Default values for channel using HDLC encoding.
DEFAULT_CHANNEL_ID = 1
DEFAULT_ADDRESS = ord('R')
STDOUT_ADDRESS = 1

FrameHandlers = dict[int, Callable[[Frame], Any]]


# Default channel output for using HDLC encoding.
def channel_output(
    writer: Callable[[bytes], Any],
    address: int = DEFAULT_ADDRESS,
    delay_s: float = 0,
) -> Callable[[bytes], None]:
    """
    Returns a function that can be used as a channel output for ``pw_rpc``.
    """

    if delay_s:

        def slow_write(data: bytes) -> None:
            """Slows down writes in case unbuffered serial is in use."""
            for byte in data:
                time.sleep(delay_s)
                writer(bytes([byte]))

        return lambda data: slow_write(encode.ui_frame(address, data))

    def write_hdlc(data: bytes):
        frame = encode.ui_frame(address, data)
        _LOG.log(_VERBOSE, 'Write %2d B: %s', len(frame), frame)
        writer(frame)

    return write_hdlc


def default_channels(write: Callable[[bytes], Any]) -> list[pw_rpc.Channel]:
    """Default Channel with HDLC encoding."""
    return [pw_rpc.Channel(DEFAULT_CHANNEL_ID, channel_output(write))]


# Writes to stdout by default, but sys.stdout.buffer is not guaranteed to exist
# (see https://docs.python.org/3/library/io.html#io.TextIOBase.buffer). Defer
# to sys.__stdout__.buffer if sys.stdout is wrapped with something that does not
# offer it.
def write_to_file(
    data: bytes,
    output: BinaryIO = getattr(sys.stdout, 'buffer', sys.__stdout__.buffer),
) -> None:
    output.write(data + b'\n')
    output.flush()


class HdlcRpcClient(client_utils.RpcClient):
    """An RPC client configured to run over HDLC.

    Expects HDLC frames to have addresses that dictate how to parse the HDLC
    payloads.
    """

    def __init__(
        self,
        reader: stream_readers.CancellableReader,
        paths_or_modules: client_utils.PathsModulesOrProtoLibrary,
        channels: Iterable[pw_rpc.Channel],
        output: Callable[[bytes], Any] = write_to_file,
        client_impl: pw_rpc.client.ClientImpl | None = None,
        *,
        _incoming_packet_filter_for_testing: (
            pw_rpc.ChannelManipulator | None
        ) = None,
        rpc_frames_address: int = DEFAULT_ADDRESS,
        log_frames_address: int = STDOUT_ADDRESS,
        extra_frame_handlers: FrameHandlers | None = None,
    ):
        """Creates an RPC client configured to communicate using HDLC.

        Args:
          reader: Readable object used to receive RPC packets.
          paths_or_modules: paths to .proto files or proto modules.
          channels: RPC channels to use for output.
          output: where to write ``stdout`` output from the device.
          client_impl: The RPC Client implementation. Defaults to the callback
            client implementation if not provided.
          rpc_frames_address: the address used in the HDLC frames for RPC
            packets. This can be the channel ID, or any custom address.
          log_frames_address: the address used in the HDLC frames for ``stdout``
            output from the device.
          extra_fram_handlers: Optional mapping of HDLC frame addresses to their
            callbacks.
        """
        # Set up frame handling.
        rpc_output: Callable[[bytes], Any] = self.handle_rpc_packet
        if _incoming_packet_filter_for_testing is not None:
            _incoming_packet_filter_for_testing.send_packet = rpc_output
            rpc_output = _incoming_packet_filter_for_testing

        frame_handlers: FrameHandlers = {
            rpc_frames_address: lambda frame: rpc_output(frame.data),
            log_frames_address: lambda frame: output(frame.data),
        }
        if extra_frame_handlers:
            frame_handlers.update(extra_frame_handlers)

        def handle_frame(frame: Frame) -> None:
            # Suppress raising any frame errors to avoid crashes on data
            # processing, which may hide or drop other data.
            try:
                if not frame.ok():
                    _LOG.error('Failed to parse frame: %s', frame.status.value)
                    _LOG.debug('%s', frame.data)
                    return

                try:
                    frame_handlers[frame.address](frame)
                except KeyError:
                    _LOG.warning(
                        'Unhandled frame for address %d: %s',
                        frame.address,
                        frame,
                    )
            except:  # pylint: disable=bare-except
                _LOG.exception('Exception in HDLC frame handler thread')

        decoder = FrameDecoder()

        def on_read_error(exc: Exception) -> None:
            _LOG.error('data reader encountered an error', exc_info=exc)

        reader_and_executor = stream_readers.DataReaderAndExecutor(
            reader, on_read_error, decoder.process_valid_frames, handle_frame
        )
        super().__init__(
            reader_and_executor, paths_or_modules, channels, client_impl
        )


class HdlcRpcLocalServerAndClient:
    """Runs an RPC server in a subprocess and connects to it over a socket.

    This can be used to run a local RPC server in an integration test.
    """

    def __init__(
        self,
        server_command: Sequence,
        port: int,
        protos: client_utils.PathsModulesOrProtoLibrary,
        *,
        incoming_processor: pw_rpc.ChannelManipulator | None = None,
        outgoing_processor: pw_rpc.ChannelManipulator | None = None,
    ) -> None:
        """Creates a new ``HdlcRpcLocalServerAndClient``."""

        self.server = stream_readers.SocketSubprocess(server_command, port)

        self._bytes_queue: queue.SimpleQueue[bytes] = queue.SimpleQueue()
        self._read_thread = threading.Thread(target=self._read_from_socket)
        self._read_thread.start()

        self.output = io.BytesIO()

        self.channel_output: Any = self.server.socket.sendall

        self._incoming_processor = incoming_processor
        if outgoing_processor is not None:
            outgoing_processor.send_packet = self.channel_output
            self.channel_output = outgoing_processor

        class QueueReader(stream_readers.CancellableReader):
            def read(self) -> bytes:
                try:
                    return self._base_obj.get(timeout=3)
                except queue.Empty:
                    return b''

            def cancel_read(self) -> None:
                pass

        self._rpc_client = HdlcRpcClient(
            QueueReader(self._bytes_queue),
            protos,
            default_channels(self.channel_output),
            self.output.write,
            _incoming_packet_filter_for_testing=incoming_processor,
        )
        self.client = self._rpc_client.client

    def _read_from_socket(self):
        while True:
            data = self.server.socket.recv(4096)
            self._bytes_queue.put(data)
            if not data:
                return

    def close(self):
        self.server.close()
        self.output.close()
        self._rpc_client.close()
        self._read_thread.join()

    def __enter__(self) -> HdlcRpcLocalServerAndClient:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()
